# coding: utf-8
"""Object-Document mapper"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import six
import collections
import shutil
import gzip

from tempfile import mkstemp, TemporaryFile, NamedTemporaryFile
from monty.json import MontyDecoder
from monty.functools import lazy_property
from mongoengine import *
from mongoengine.fields import GridFSProxy
from mongoengine.base.datastructures import BaseDict
from abipy import abilab

import logging
logger = logging.getLogger(__name__)


class AbiGridFSProxy(GridFSProxy):

    def abiopen(self):
        """Dump the gridfs data to a temporary file and use ``abiopen`` to open the file."""
        _, filepath = mkstemp(suffix='.' + self.abiext, text=self.abiform == "t")

        with open(filepath , "w" + self.abiform) as fh:
            fh.write(self.read())

        return abilab.abiopen(filepath)


class AbiFileField(FileField):
    """
    Extend ``FileField``. Use customized version of proxy_class so that
    we can use ``abiopen`` to construct the AbiPy object from the gridfs content.
    """
    proxy_class = AbiGridFSProxy

    def __init__(self, **kwargs):
        self.abiext = kwargs.pop("abiext")
        self.abiform = kwargs.pop("abiform")

        super(AbiFileField, self).__init__(**kwargs)

    def _monkey_patch_proxy(self, proxy):
        """
        Monkey patch the proxy adding ``abiext`` and ``abiform``.
        so that we know how to open the file with ``abiopen``.
        """
        proxy.abiext, proxy.abiform = self.abiext, self.abiform
        return proxy

    def get_proxy_obj(self, **kwargs):
        proxy = super(AbiFileField, self).get_proxy_obj(**kwargs)
        return self._monkey_patch_proxy(proxy)

    def to_python(self, value):
        if value is not None:
            proxy = super(AbiFileField, self).to_python(value)
            return self._monkey_patch_proxy(proxy)


class GzipGridFSProxy(GridFSProxy):
    """
    Proxy object to handle writing and reading of files to and from GridFS.
    Files are compressed with gzip before being saved in the GridFS.
    To decompress the object exposes an unzip method.
    """

    def put(self, file_obj, **kwargs):
        # try to check if the file is already zipped.
        # in that case just handle as a normal file
        try:
            import magic
            # check that the module is actually python-magic and not libmagic
            f = getattr(magic, "from_buffer", None)
            if f and callable(f):
                if magic.from_buffer(file_obj.read(3), mime=True) == "application/gzip":
                    file_obj.seek(0)
                    return super(GzipGridFSProxy, self).put(file_obj, **kwargs)
                        
            file_obj.seek(0)

        except ImportError:
            logger.info("No python-magic library available. Skipping check...")

        field = self.instance._fields[self.key]
        # Handle nested fields
        if hasattr(field, 'field') and isinstance(field.field, FileField):
            field = field.field

        with TemporaryFile() as tmp_file:
            tmp_gz = gzip.GzipFile(fileobj=tmp_file, compresslevel=field.compresslevel, mode='w+b')
            shutil.copyfileobj(file_obj, tmp_gz)
            tmp_gz.close()
            tmp_file.seek(0)

            return super(GzipGridFSProxy, self).put(tmp_file, **kwargs)

    def write(self, *args, **kwargs):
        raise RuntimeError("Please use \"put\" method instead")

    def writelines(self, *args, **kwargs):
        raise RuntimeError("Please use \"put\" method instead")

    def to_gzip(self):
        return gzip.GzipFile(fileobj=self.get())

    def unzip(self, filepath=None, mode='w+b'):

        with open(filepath, mode) as f:
            f.write(self.to_gzip().read())
            return f


class GzipFileField(FileField):
    """
    A GridFS storage field with automatic compression of the file with gzip.
    """

    proxy_class = GzipGridFSProxy

    def __init__(self, compresslevel=9, **kwargs):
        self.compresslevel = compresslevel
        super(GzipFileField, self).__init__(**kwargs)


class AbiGzipFSProxy(GzipGridFSProxy):
    """
    Proxy object to handle writing and reading of abinit related files to and from GridFS.
    Files are compressed with gzip before being saved in the GridFS.
    To decompress the object exposes an unzip method.
    """

    def abiopen(self):
        """Dump the unzipped gridfs data to a temporary file and use `abiopen` to open the file."""
        field = self.instance._fields[self.key]
        # Handle nested fields
        if hasattr(field, 'field') and isinstance(field.field, FileField):
            field = field.field

        mode = 'w+'
        if field.abiform == 'b':
            mode += 'b'
        with NamedTemporaryFile(suffix='.' + field.abiext, mode=mode) as tmp_abifile:
            tmp_abifile.write(self.to_gzip().read())
            return abilab.abiopen(tmp_abifile.name)


class AbiGzipFileField(GzipFileField):
    """
    A GridFS storage field for abinit related files with automatic compression of the file with gzip.
    """

    proxy_class = AbiGzipFSProxy

    def __init__(self, abiext, abiform, compresslevel=9, **kwargs):
        self.abiext =abiext
        self.abiform = abiform
        super(AbiGzipFileField, self).__init__(compresslevel=compresslevel, **kwargs)


class MongoFiles(EmbeddedDocument):
    """
    Document with the output files produced by the AbiPy |Task|
    (references to GridFs files)
    """
    gsr = AbiFileField(abiext="GSR.nc", abiform="b")
    hist = AbiFileField(abiext="HIST", abiform="b")
    phbst = AbiFileField(abiext="PHBST.nc", abiform="b")
    phdos = AbiFileField(abiext="PHDOS.nc", abiform="b")
    sigres = AbiFileField(abiext="SIGRES.nc", abiform="b")
    mdf = AbiFileField(abiext="MDF.nc", abiform="b")

    ddb = AbiFileField(abiext="DDB", abiform="t")
    output_file = AbiFileField(abiext="abo", abiform="t")

    @classmethod
    def from_node(cls, node):
        """Add to GridFs the files produced in the ``outdir`` of the node."""
        new = cls()

        for key, field in cls._fields.items():
            if not isinstance(field, AbiFileField): continue
            ext, form = field.abiext, field.abiform

            path = node.outdir.has_abiext(ext)
            if path:
                with open(path, "r" + form) as f:
                    # Example: new.gsr.put(f)
                    fname = ext.replace(".nc", "").lower()
                    proxy = getattr(new, fname)
                    proxy.put(f)

        # Special treatment of the main output
        # (the file is not located in node.outdir)
        if hasattr(node, "output_file"):
            #print("in out")
            new.output_file.put(node.output_file.read())

        return new

    def delete(self):
        """Delete gridFs files"""
        for field in self._fields.values():
            if not isinstance(field, AbiFileField): continue
            value = getattr(self, field.name)
            if hasattr(value, "delete"):
                print("Deleting %s" % field.name)
                value.delete()


class MSONDict(BaseDict):
    def to_mgobj(self):
        return MontyDecoder().process_decoded(self)


class MSONField(DictField):

    def __get__(self, instance, owner):
        """Descriptor for retrieving a value from a field in a document.
        """
        value = super(MSONField, self).__get__(instance, owner)
        if isinstance(value, BaseDict):
            value.__class__ = MSONDict

        # print("value:", type(value))
        return value

    #def to_python(self, value):
    #    #print("to value:", type(value))
    #    if isinstance(value, collections.Mapping) and "@module" in value:
    #        value = MontyDecoder().process_decoded(value)
    #    else:
    #        value = super(MSONField, self).to_python(value)

    #    print("to value:", type(value))
    #    return value

    #def to_mongo(self, value):
    #    #print(value.as_dict())
    #    return value.as_dict()


class MongoTaskResults(EmbeddedDocument):
    """Document with the most important results produced by the AbiPy |Task|"""

    #meta = {'allow_inheritance': True}

    #: The initial input structure for the calculation in the pymatgen json representation
    #initial_structure = DictField(required=True)
    initial_structure = MSONField(required=True)

    #: The final relaxed structure in a dict format.
    final_structure = DictField(required=True)

    @classmethod
    def from_task(cls, task):
        # TODO Different Documents depending on task.__class__ or duck typing?
        #initial_structure = MSONField().to_mongo(task.input.structure.as_dict())
        initial_structure = task.input.structure.as_dict()
        #print(type(initial_structure))
        final_structure = initial_structure

        if hasattr(task, "open_gsr"):
            with task.open_gsr() as gsr:
                final_structure = gsr.structure.as_dict()

        new = cls(
            initial_structure=initial_structure,
            final_structure=final_structure,
        )
        return new


class MongoNode(Document):

    meta = {
        'allow_inheritance': True,
        "collection": "flowdata",
    }
    #meta = {'meta': True}

    node_id = LongField(required=True)
    node_class = StringField(required=True)
    status = StringField(required=True)
    workdir = StringField(required=True)

    #date_modified = DateTimeField(default=datetime.datetime.now)

    @classmethod
    def from_node(cls, node):
        return cls(
            node_class=node.__class__.__name__,
            node_id=node.node_id,
            status=str(node.status),
            workdir=node.workdir,
        )


class MongoEmbeddedNode(EmbeddedDocument):

    meta = {'allow_inheritance': True}
    #meta = {'meta': True}

    node_id = LongField(required=True)

    node_class = StringField(required=True)

    status = StringField(required=True)

    workdir = StringField(required=True)

    #date_modified = DateTimeField(default=datetime.datetime.now)

    @classmethod
    def from_node(cls, node):
        return cls(
            node_class=node.__class__.__name__,
            node_id=node.node_id,
            status=str(node.status),
            workdir=node.workdir,
        )


class MongoTask(MongoEmbeddedNode):
    """Document associated to an AbiPy |Task|"""

    input = DictField(required=True)
    input_str = StringField(required=True)

    # Abinit events.
    report = DictField(required=True)
    num_warnings = IntField(required=True, help_text="Number of warnings")
    num_errors = IntField(required=True, help_text="Number of errors")
    num_comments =  IntField(required=True, help_text="Number of comments")

    #: Total CPU time taken.
    #cpu_time = FloatField(required=True)
    #: Total wall time taken.
    #wall_time = FloatField(required=True)

    results = EmbeddedDocumentField(MongoTaskResults)

    outfiles = EmbeddedDocumentField(MongoFiles)

    #@property
    #def num_warnings(self):
    #    """Number of warnings reported."""
    #    return self.input.num_warnings

    #@property
    #def num_errors(self):
    #    """Number of errors reported."""
    #    return self.input.num_error

    #@property
    #def num_comments(self):
    #    """Number of comments reported."""
    #    return len(self.comments)

    #@property
    #def is_paw(self):
    #    print("in is paw")
    #    return True

    @classmethod
    def from_task(cls, task):
        """Build the document from an AbiPy |Task| object."""
        new = cls.from_node(task)

        new.input = task.input.as_dict()
        new.input_str = str(task.input)

        # TODO: Handle None!
        report = task.get_event_report()
        for a in ("num_errors", "num_comments", "num_warnings"):
            setattr(new, a, getattr(report, a))
        new.report = report.as_dict()

        new.results = MongoTaskResults.from_task(task)
        new.outfiles = MongoFiles.from_node(task)

        return new


class MongoWork(MongoEmbeddedNode):
    """Document associated to an AbiPy |Work|"""

    #: List of tasks.
    tasks = ListField(EmbeddedDocumentField(MongoTask), required=True)

    #: Output files produced by the work.
    outfiles = EmbeddedDocumentField(MongoFiles)

    @classmethod
    def from_work(cls, work):
        """Build and return the document from a AbiPy |Work| instance."""
        new = cls.from_node(work)
        new.tasks = [MongoTask.from_task(task) for task in work]
        new.outfiles = MongoFiles.from_node(work)
        return new

    def __getitem__(self, name):
        try:
            # Dictionary-style field of super
            return super(MongoWork, self).__getitem__(name)
        except KeyError:
            # Assume int or slice
            try:
                return self.tasks[name]
            except IndexError:
                raise


class MongoFlow(MongoNode):
    """
    Document associated to an AbiPy |Flow|

    Assumptions:
        All the tasks must have the same list of pseudos,
        same chemical formula.
    """

    #: List of works
    works = ListField(EmbeddedDocumentField(MongoWork), required=True)

    #: Output files produced by the flow.
    outfiles = EmbeddedDocumentField(MongoFiles)

    #meta = {
    #    "collection": "flowdata",
    #    #"indexes": ["status", "priority", "created"],
    #}

    @classmethod
    def from_flow(cls, flow):
        """Build and return the document from a AbiPy |Flow| instance."""
        new = cls.from_node(flow)
        new.works = [MongoWork.from_work(work) for work in flow]
        new.outfiles = MongoFiles.from_node(flow)
        #new.assimilated = flow.mongo_assimilate()
        return new

    def __getitem__(self, name):
        try:
            # Dictionary-style field of super
            return super(MongoFlow, self).__getitem__(name)
        except KeyError:
            # Assume int or slice
            try:
                return self.works[name]
            except IndexError:
                raise

    def pickle_load(self):
        """
        Load the pickle file from the working directory of the flow.

        Return: AbiPy |Flow| object.
        """
        flow = abilab.Flow.pickle_load(self.workdir)
        #flow.set_mongo_id(self.id)
        return flow

    def delete(self):
        # Remove GridFs files.
        for work in self.works:
            work.outfiles.delete()
            #work.delete()
            for task in work:
                #task.delete()
                task.outfiles.delete()

        self.delete()

    @queryset_manager
    def completed(doc_cls, queryset):
        return queryset.filter(status="Completed")

    #@queryset_manager
    #def running(doc_cls, queryset):
    #    return queryset.filter(status__in=["AbiCritical", "QCritical", "Error",])

    #@queryset_manager
    #def paw_flows(doc_cls, queryset):
    #    return queryset.filter(is_paw=True)

    #@queryset_manager
    #def nc_flows(doc_cls, queryset):
    #    return queryset.filter(is_nc=True)



from __future__ import print_function, division, unicode_literals, absolute_import

import numpy as np
import os


from pymatgen.io.vasp.sets import MPRelaxSet
from pymatgen.io.vasp.outputs import Poscar
from pymatgen.core.structure import Structure
from pymatgen.core.sites import PeriodicSite
from itertools import chain


class MPNEBSet(MPRelaxSet):

    def __init__(self, structures, unset_encut=False, **kwargs):
        if len(structures) < 3:
            raise ValueError("You need at least 3 structures for an NEB.")
        kwargs["sort_structure"] = False
        MPRelaxSet.__init__(self, structures[0], **kwargs)
        self.structures = self._process_structures(structures)
        self.unset_encut = False
        if unset_encut:
            self.config_dict["INCAR"].pop("ENCUT", None)

        if "EDIFF" not in self.config_dict["INCAR"]:
            self.config_dict["INCAR"]["EDIFF"] = self.config_dict[
                "INCAR"].pop("EDIFF_PER_ATOM")

        # NEB specific defaults
        defaults = {'IMAGES': len(structures) - 2, 'IBRION': 1, 'ISYM': 0,
                    'LCHARG': False, "LDAU": False}
        self.config_dict["INCAR"].update(defaults)

    @property
    def poscar(self):
        return Poscar(self.structures[0])

    @property
    def poscars(self):
        return [Poscar(s) for s in self.structures]

    def _process_structures(self, structures):
        """
        Remove any atom jumps across the cell
        """
        input_structures = structures
        structures = [input_structures[0]]
        for s in input_structures[1:]:
            prev = structures[-1]
            for i in range(len(s)):
                t = np.round(prev[i].frac_coords - s[i].frac_coords)
                if np.sum(t) > 0.5:
                    s.translate_sites([i], t, to_unit_cell=False)
            structures.append(s)
        return structures

    def write_input(self, output_dir, make_dir_if_not_present=True,
                    write_cif=False, write_path_cif=False,
                    write_endpoint_inputs=False):
        """
        NEB inputs has a special directory structure where inputs are in 00,
        01, 02, ....

        Args:
            output_dir (str): Directory to output the VASP input files
            make_dir_if_not_present (bool): Set to True if you want the
                directory (and the whole path) to be created if it is not
                present.
            write_cif (bool): If true, writes a cif along with each POSCAR.
            write_path_cif (bool): If true, writes a cif for each image.
            write_endpoint_inputs (bool): If true, writes input files for
                running endpoint calculations.
        """

        if make_dir_if_not_present and not os.path.exists(output_dir):
            os.makedirs(output_dir)
        self.incar.write_file(os.path.join(output_dir, 'INCAR'))
        self.kpoints.write_file(os.path.join(output_dir, 'KPOINTS'))
        self.potcar.write_file(os.path.join(output_dir, 'POTCAR'))

        for i, p in enumerate(self.poscars):
            d = os.path.join(output_dir, str(i).zfill(2))
            if not os.path.exists(d):
                os.makedirs(d)
            p.write_file(os.path.join(d, 'POSCAR'))
            if write_cif:
                p.structure.to(filename=os.path.join(d, '{}.cif'.format(i)))
        if write_endpoint_inputs:
            end_point_param = MPRelaxSet(
                self.structures[0],
                user_incar_settings=self.user_incar_settings)

            for image in ['00', str(len(self.structures) - 1).zfill(2)]:
                end_point_param.incar.write_file(os.path.join(output_dir, image, 'INCAR'))
                end_point_param.kpoints.write_file(os.path.join(output_dir, image, 'KPOINTS'))
                end_point_param.potcar.write_file(os.path.join(output_dir, image, 'POTCAR'))
        if write_path_cif:
            sites = set()
            l = self.structures[0].lattice
            for site in chain(*(s.sites for s in self.structures)):
                sites.add(PeriodicSite(site.species_and_occu, site.frac_coords, l))
            path = Structure.from_sites(sorted(sites))
            path.to(filename=os.path.join(output_dir, 'path.cif'))


class MPcNEBSet(MPNEBSet):
    def __init__(self, structures, unset_encut=False, **kwargs):
        super(MPcNEBSet, self).__init__(structures, unset_encut=False, **kwargs)
        # cNEB specific defaults
        defaults = {'LCLIMB': True}
        self.config_dict["INCAR"].update(defaults)


from __future__ import print_function, division, unicode_literals, absolute_import

import logging
import json

from fireworks.core.firework import FireTaskBase
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometry_finder import LocalGeometryFinder
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import LightStructureEnvironments
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import StructureEnvironments
from pymatgen.analysis.bond_valence import BVAnalyzer
try:
    from pymatgen.ext.matproj import MPRester, MPRestError
except ImportError:
    from pymatgen.matproj.rest import MPRester, MPRestError


class ChemEnvStructureEnvironmentsTask(FireTaskBase):
    _fw_name = "ChemEnvStructureEnvironmentsTask"

    def run_task(self, fw_spec):
        logging.basicConfig(filename='chemenv_structure_environments.log',
                            format='%(levelname)s:%(module)s:%(funcName)s:%(message)s',
                            level=logging.DEBUG)
        lgf = LocalGeometryFinder()
        lgf.setup_parameters(centering_type='centroid', include_central_site_in_centroid=True,
                             structure_refinement=lgf.STRUCTURE_REFINEMENT_NONE)
        if 'chemenv_parameters' in fw_spec:
            for param, value in fw_spec['chemenv_parameters'].items():
                lgf.setup_parameter(param, value)
        identifier = fw_spec['identifier']
        if 'structure' in fw_spec:
            structure = fw_spec['structure']
        else:
            if identifier['source'] == 'MaterialsProject' and 'material_id' in identifier:
                if not 'mapi_key' in fw_spec:
                    raise ValueError('The mapi_key should be provided to get the structure from the Materials Project')
                a = MPRester(fw_spec['mapi_key'])
                structure = a.get_structure_by_material_id(identifier['material_id'])
            else:
                raise ValueError('Either structure or identifier with source = MaterialsProject and material_id '
                                 'should be provided')

        info = {}
        # Compute the structure environments
        lgf.setup_structure(structure)
        if 'valences' in fw_spec:
            valences = fw_spec['valences']
        else:
            try:
                bva = BVAnalyzer()
                valences = bva.get_valences(structure=structure)
                info['valences'] = {'origin': 'BVAnalyzer'}
            except:
                valences = 'undefined'
                info['valences'] = {'origin': 'None'}
        excluded_atoms = None
        if 'excluded_atoms' in fw_spec:
            excluded_atoms = fw_spec['excluded_atoms']

        se = lgf.compute_structure_environments(only_cations=False, valences=valences, excluded_atoms=excluded_atoms)

        # Write to json file
        if 'json_file' in fw_spec:
            json_file = fw_spec['json_file']
        else:
            json_file = 'structure_environments.json'
        f = open(json_file, 'w')
        json.dump(se.as_dict(), f)
        f.close()

        # Save to database
        if 'mongo_database' in fw_spec:
            database = fw_spec['mongo_database']
            entry = {'identifier': identifier,
                     'elements': [elmt.symbol for elmt in structure.composition.elements],
                     'nelements': len(structure.composition.elements),
                     'pretty_formula': structure.composition.reduced_formula,
                     'nsites': len(structure)
                     }

            saving_option = fw_spec['saving_option']
            if saving_option == 'gridfs':
                gridfs_msonables = {'structure': structure,
                                    'structure_environments': se}
            elif saving_option == 'storefile':
                gridfs_msonables = None
                if 'se_prefix' in fw_spec:
                    se_prefix = fw_spec['se_prefix']
                    if not se_prefix.isalpha():
                        raise ValueError('Prefix for structure_environments file is "{}" '
                                         'while it should be alphabetic'.format(se_prefix))
                else:
                    se_prefix = ''
                if se_prefix:
                    se_rfilename = '{}_{}.json'.format(se_prefix, fw_spec['storefile_basename'])
                else:
                    se_rfilename = '{}.json'.format(fw_spec['storefile_basename'])
                se_rfilepath = '{}/{}'.format(fw_spec['storefile_dirpath'], se_rfilename)
                storage_server = fw_spec['storage_server']
                storage_server.put(localpath=json_file, remotepath=se_rfilepath, overwrite=True, makedirs=False)
                entry['structure_environments_file'] = se_rfilepath
            else:
                raise ValueError('Saving option is "{}" while it should be '
                                 '"gridfs" or "storefile"'.format(saving_option))
            criteria = {'identifier': identifier}
            if database.collection.find(criteria).count() == 1:
                database.update_entry(query=criteria, entry_update=entry,
                                      gridfs_msonables=gridfs_msonables)
            else:
                database.insert_entry(entry=entry, gridfs_msonables=gridfs_msonables)

class ChemEnvLightStructureEnvironmentsTask(FireTaskBase):
    _fw_name = "ChemEnvLightStructureEnvironmentsTask"

    def run_task(self, fw_spec):
        logging.basicConfig(filename='chemenv_light_structure_environments.log',
                            format='%(levelname)s:%(module)s:%(funcName)s:%(message)s',
                            level=logging.DEBUG)

        identifier = fw_spec['identifier']
        criteria = {'identifier': identifier}
        # Where to get the full structure environments object
        se_database = fw_spec['structure_environments_database']
        entry = se_database.collection.find_one(criteria)
        if fw_spec['structure_environments_setup'] == 'from_gridfs':
            gfs_fileobject = se_database.gridfs.get(entry['structure_environments'])
            dd = json.load(gfs_fileobject)
            se = StructureEnvironments.from_dict(dd)
        elif fw_spec['structure_environments_setup'] == 'from_storedfile':
            se_storage_server = fw_spec['se_storage_server']
            se_filepath = entry['structure_environments_file']
            se_storage_server.get(se_filepath, 'se.json')
            f = open('se.json', 'r')
            dd = json.load(f)
            f.close()
            se = StructureEnvironments.from_dict(dd)
        else:
            raise RuntimeError('Wrong structure_environments_setup : '
                               '"{}" is not allowed'.format(fw_spec['structure_environments_setup']))

        # Compute the light structure environments
        chemenv_strategy = fw_spec['chemenv_strategy']
        if 'valences' in fw_spec:
            valences = fw_spec['valences']
            valences_origin = fw_spec['valences_origin']
        else:
            valences = 'undefined'
            valences_origin = 'None'
        lse = LightStructureEnvironments.from_structure_environments(strategy=chemenv_strategy,
                                                                     structure_environments=se,
                                                                     valences=valences,
                                                                     valences_origin=valences_origin)

        # Write to json file
        if 'json_file' in fw_spec:
            json_file = fw_spec['json_file']
        else:
            json_file = 'light_structure_environments.json'
        f = open(json_file, 'w')
        json.dump(lse.as_dict(), f)
        f.close()

        # Save to database
        if 'mongo_database' in fw_spec:
            database = fw_spec['mongo_database']
            entry = {'identifier': identifier,
                     'elements': [elmt.symbol for elmt in lse.structure.composition.elements],
                     'nelements': len(lse.structure.composition.elements),
                     'pretty_formula': lse.structure.composition.reduced_formula,
                     'nsites': len(lse.structure),
                     'chemenv_statistics': lse.get_statistics(statistics_fields='ALL', bson_compatible=True)
                     }
            saving_option = fw_spec['saving_option']
            if saving_option == 'gridfs':
                gridfs_msonables = {'structure': lse.structure,
                                    'light_structure_environments': lse}
            elif saving_option == 'storefile':
                gridfs_msonables = None
                if 'lse_prefix' in fw_spec:
                    lse_prefix = fw_spec['lse_prefix']
                    if not lse_prefix.isalpha():
                        raise ValueError('Prefix for light_structure_environments file is "{}" '
                                         'while it should be alphabetic'.format(lse_prefix))
                else:
                    lse_prefix = ''
                if lse_prefix:
                    lse_rfilename = '{}_{}.json'.format(lse_prefix, fw_spec['storefile_basename'])
                else:
                    lse_rfilename = '{}.json'.format(fw_spec['storefile_basename'])
                lse_rfilepath = '{}/{}'.format(fw_spec['storefile_dirpath'], lse_rfilename)
                storage_server = fw_spec['storage_server']
                storage_server.put(localpath=json_file, remotepath=lse_rfilepath, overwrite=True, makedirs=False)
                entry['light_structure_environments_file'] = lse_rfilepath
            else:
                raise ValueError('Saving option is "{}" while it should be '
                                 '"gridfs" or "storefile"'.format(saving_option))
            criteria = {'identifier': identifier}
            if database.collection.find(criteria).count() == 1:
                database.update_entry(query=criteria, entry_update=entry,
                                      gridfs_msonables=gridfs_msonables)
            else:
                database.insert_entry(entry=entry, gridfs_msonables=gridfs_msonables)


# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import shutil

from abiflows.core.mastermind_abc import Cleaner
from abiflows.core.testing import AbiflowsTest


test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..",
                        "test_files")

class TestCleaner(AbiflowsTest):

    def test_cleaner(self):
        # Keep current working directory, create tmp directory and change to tmp directory
        cwd = os.getcwd()
        tmp_dir = 'tmp'
        os.makedirs(tmp_dir)
        os.chdir(tmp_dir)

        tmp_abs_dir = os.getcwd()

        # Create a list of files and directories
        os.makedirs('outputs')
        os.makedirs('outputs/formatted')
        os.makedirs('outputs/text')
        os.makedirs('results')
        os.makedirs('temporary')

        os.mknod('somefile.txt')
        os.mknod('somefile.txt.backup')

        os.mknod('outputs/text/text1.abc')
        os.mknod('outputs/text/text2.abc')
        os.mknod('outputs/text/text3.abc')
        os.mknod('outputs/text/text1.def')
        os.mknod('outputs/text/text15.def')

        os.mknod('outputs/formatted/formatted1.txt')
        os.mknod('outputs/formatted/formatted2.txt')
        os.mknod('outputs/formatted/formatted3.log')
        os.mknod('outputs/formatted/formatted4.log')
        os.mknod('outputs/formatted/formatted5.log')
        os.mknod('outputs/formatted/formatted6.bin')
        os.mknod('outputs/formatted/formatted7.bog')
        os.mknod('outputs/formatted/formatted8.beg')

        os.mknod('temporary/item.log')
        os.mknod('temporary/result.txt')

        # Create a first cleaner
        cleaner1 = Cleaner(dirs_and_patterns=[{'directory': 'outputs/text',
                                               'patterns': ['text?.abc']}])
        cleaner1.clean(root_directory=tmp_abs_dir)

        # Check that the first cleaner did his job correctly
        self.assertTrue(os.path.exists('outputs/text/text1.def'))
        self.assertTrue(os.path.exists('outputs/text/text15.def'))
        self.assertFalse(os.path.exists('outputs/text/text1.abc'))
        self.assertFalse(os.path.exists('outputs/text/text2.abc'))
        self.assertFalse(os.path.exists('outputs/text/text3.abc'))

        # Create a second cleaner
        cleaner2 = Cleaner(dirs_and_patterns=[{'directory': '.',
                                               'patterns': ['temporary']},
                                              {'directory': 'outputs/formatted',
                                               'patterns': ['*[1-4].log', '*.b?g']}])
        cleaner2.clean(root_directory=tmp_abs_dir)

        # Check that the first cleaner did his job correctly
        self.assertTrue(os.path.exists('outputs/formatted/formatted1.txt'))
        self.assertTrue(os.path.exists('outputs/formatted/formatted2.txt'))
        self.assertFalse(os.path.exists('outputs/formatted/formatted3.log'))
        self.assertFalse(os.path.exists('outputs/formatted/formatted4.log'))
        self.assertTrue(os.path.exists('outputs/formatted/formatted5.log'))
        self.assertTrue(os.path.exists('outputs/formatted/formatted6.bin'))
        self.assertFalse(os.path.exists('outputs/formatted/formatted7.bog'))
        self.assertFalse(os.path.exists('outputs/formatted/formatted8.beg'))
        self.assertFalse(os.path.exists('temporary'))
        self.assertTrue(os.path.exists('outputs/formatted'))

        # Change back to the initial working directory and remove the tmp directory
        os.chdir(cwd)
        shutil.rmtree(tmp_dir)



# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import abipy.data as abidata
import abipy.abilab as abilab

from abiflows.fireworks.tasks.abinit_tasks import RestartInfo, AbinitRuntimeError, RelaxFWTask
from abiflows.fireworks.utils.task_history import TaskHistory, TaskEvent
from abiflows.core.testing import AbiflowsTest
from abipy.abio.factories import ion_ioncell_relax_input
from abipy.flowtk import events# Correction, DilatmxErrorHandler, DilatmxError


class TestTaskHistory(AbiflowsTest):

    def test_task_history_and_events(self):
        si = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        si_relax_input = ion_ioncell_relax_input(si, abidata.pseudos("14si.pspnc"), ecut=2).split_datasets()[1]

        th = TaskHistory()
        th.log_initialization(task=RelaxFWTask(si_relax_input), initialization_info={"test_info": 1})
        th.log_autoparal({u'time': u'12:0:0', u'ntasks': 15, u'partition': 'defq', u'nodes': 1, u'mem_per_cpu': 3000})
        th.log_finalized(final_input=si_relax_input)
        th.log_restart(RestartInfo(os.path.abspath('.'), reset=True, num_restarts=2))
        th.log_unconverged()
        th.log_corrections([events.Correction(events.DilatmxErrorHandler(), {}, events.DilatmxError('', '', '',), )])
        th.log_abinit_stop(run_time=100)

        try:
            raise AbinitRuntimeError("test error")
        except AbinitRuntimeError as exc:
            th.log_error(exc)

        try:
            raise RuntimeError("test error")
        except RuntimeError as exc:
            th.log_error(exc)

        th.log_converge_params({'dilatmx': 1.03}, si_relax_input)

        self.assertMSONable(th)

        for te in th:
            self.assertMSONable(te)

    def test_get_events_by_types(self):
        th = TaskHistory()
        th.log_unconverged()
        th.log_unconverged()
        th.log_abinit_stop(run_time=100)
        th.log_finalized()

        events = th.get_events_by_types([TaskEvent.UNCONVERGED, TaskEvent.FINALIZED])

        self.assertEqual(len(th), 4)
        self.assertEqual(len(events), 3)

    def test_get_total_run_time(self):
        th = TaskHistory()
        th.log_abinit_stop(run_time=100)
        th.log_abinit_stop(run_time=100)
        th.log_abinit_stop(run_time=100)

        total_run_time = th.get_total_run_time()

        self.assertEqual(total_run_time, 300)





# coding: utf-8
"""
Task history related objects
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import collections
import traceback
import logging

from monty.json import MontyDecoder, jsanitize, MSONable
from pymatgen.util.serialization import pmg_serialize


logger = logging.getLogger(__name__)


class TaskHistory(collections.deque, MSONable):
    """
    History class for tracking the creation and actions performed during a task.
    The objects provided should be PGMSONable, thus dicts, lists or PGMSONable objects.
    The expected items are dictionaries, transformations, corrections, autoparal, restart/reset
    and initializations (when factories will be handled).
    Possibly, the first item should contain information about the starting point of the task.
    This object will be forwarded during task restarts and resets, in order to keep track of the full history of the
    task.
    """

    @pmg_serialize
    def as_dict(self):
        items = [i.as_dict() if hasattr(i, "as_dict") else i for i in self]
        return dict(items=items)

    @classmethod
    def from_dict(cls, d):
        dec = MontyDecoder()
        return cls([dec.process_decoded(i) for i in d['items']])

    def log_initialization(self, task, initialization_info=None):
        details = {'task_class': task.__class__.__name__}
        if initialization_info:
            details['initialization_info'] = initialization_info
        self.append(TaskEvent(TaskEvent.INITIALIZED, details=details))

    def log_corrections(self, corrections):
        self.append(TaskEvent(TaskEvent.CORRECTIONS, corrections))

    def log_restart(self, restart_info, local_restart=False):
        self.append(TaskEvent(TaskEvent.RESTART, details=dict(restart_info=restart_info, local_restart=local_restart)))

    def log_autoparal(self, optconf):
        self.append(TaskEvent(TaskEvent.AUTOPARAL, details={'optconf': optconf}))

    def log_unconverged(self):
        self.append(TaskEvent(TaskEvent.UNCONVERGED))

    def log_finalized(self, final_input=None):
        details = dict(total_run_time=self.get_total_run_time())
        if final_input:
            details['final_input'] = final_input
        self.append(TaskEvent(TaskEvent.FINALIZED, details=details))

    def log_converge_params(self, unconverged_params, abiinput):
        params={}
        for param, new_value in unconverged_params.items():
            params[param] = dict(old_value=abiinput.get(param, 'Default'), new_value=new_value)
        self.append(TaskEvent(TaskEvent.UNCONVERGED_PARAMS, details={'params': params}))

    def log_error(self, exc):
        tb = traceback.format_exc()
        event_details = dict(stacktrace=tb)
        # If the exception is serializable, save its details
        try:
            exception_details = exc.to_dict()
        except AttributeError:
            exception_details = None
        except BaseException as e:
            logger.error("Exception couldn't be serialized: {} ".format(e))
            exception_details = None
        if exception_details:
            event_details['exception_details'] = exception_details
        self.append(TaskEvent(TaskEvent.ERROR, details=event_details))

    def log_abinit_stop(self, run_time=None):
        self.append(TaskEvent(TaskEvent.ABINIT_STOP, details={'run_time': run_time}))


    def get_events_by_types(self, types):
        """
        Return the events in history of the selected types. types can be a single type or a list
        """

        types = types if isinstance(types, (list, tuple)) else [types]

        events = [e for e in self if e.event_type in types]

        return events

    def get_total_run_time(self):
        """
        Calculates total run time based summing the run times saved in the abinit stop event.
        """

        total_run_time = 0
        for te in self.get_events_by_types(TaskEvent.ABINIT_STOP):
            run_time = te.details.get('run_time', None)
            if run_time:
                total_run_time += run_time

        return total_run_time


class TaskEvent(MSONable):
    """
    Object used to categorize the events in the TaskHistory.
    """

    INITIALIZED = 'initialized'
    CORRECTIONS = 'corrections'
    RESTART = 'restart'
    AUTOPARAL = 'autoparal'
    UNCONVERGED = 'unconverged'
    FINALIZED = 'finalized'
    UNCONVERGED_PARAMS = 'unconverged parameters'
    ERROR = 'error'
    ABINIT_STOP = 'abinit stop'

    def __init__(self, event_type, details=None):
        self.event_type = event_type
        self.details = details

    @pmg_serialize
    def as_dict(self):
        d = dict(event_type=self.event_type)
        if self.details:
            d['details'] = jsanitize(self.details, strict=True)

        return d

    @classmethod
    def from_dict(cls, d):
        dec = MontyDecoder()
        details = dec.process_decoded(d['details']) if 'details' in d else None
        return cls(event_type=d['event_type'], details=details)




# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import abipy.data as abidata
import abipy.abilab as abilab
import abiflows.fireworks.tasks.abinit_tasks as abinit_tasks
import abiflows.fireworks.utils.fw_utils

from abipy.abio.factories import *
from abipy.abio.factories import ScfForPhononsFactory
from abipy.abio.inputs import AnaddbInput
from abiflows.core.testing import AbiflowsTest
from abiflows.fireworks.tasks.tests import mock_objects
from fireworks import FWAction


mock = AbiflowsTest.get_mock_module()


test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..",
                        "test_files", "fw_task_managers")


class TestAbiFireTask(AbiflowsTest):

    def setUp(self):
        self.si_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        self.si_scf_input = ebands_input(self.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10).split_datasets()[0]
        self.si_scf_factory = ScfForPhononsFactory(self.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10)

    def test_AbiFireTask_basic(self):
        task = abinit_tasks.AbiFireTask(self.si_scf_input)
        task.to_dict()
        self.assertMSONable(self.si_scf_input)
        self.assertFwSerializable(task)

    def test_AbiFireTask_methods(self):
        task = abinit_tasks.AbiFireTask(self.si_scf_input)

        assert task.get_fw_task_manager({'ftm_file': os.path.join(test_dir, "fw_manager_ok.yaml")})

        ftm_path = os.path.join(test_dir, "fw_manager_ok.yaml")
        ftm = abiflows.fireworks.utils.fw_utils.FWTaskManager.from_file(ftm_path)
        # remove the task manager to make the function fail
        ftm.task_manager = None
        with self.assertRaises(abinit_tasks.InitializationError):
            task.run_autoparal(self.si_scf_input, '.', ftm)

        ftm_path = os.path.join(test_dir, "fw_manager_walltime_command.yaml")
        task.is_autoparal = None
        task.setup_task({'ftm_file': ftm_path})
        self.assertEqual(task.walltime, 1000)


    def test_ScfFireTask(self):
        task = abinit_tasks.ScfFWTask(self.si_scf_input)
        task.to_dict()
        self.assertFwSerializable(task)

        task = abinit_tasks.ScfFWTask(self.si_scf_factory)
        task.to_dict()
        self.assertFwSerializable(task)

    def test_AnaDdbAbinitTask(self):
        ana_inp = AnaddbInput.phbands_and_dos(self.si_structure, ngqpt=[4,4,4], nqsmall=10)
        task = abinit_tasks.AnaDdbAbinitTask(ana_inp)
        task.to_dict()
        self.assertFwSerializable(task)


class TestTaskAnalysis(AbiflowsTest):
    def setUp(self):
        self.si_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        self.si_scf_input = ebands_input(self.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10).split_datasets()[0]

    @mock.patch.object(abinit_tasks.AbiFireTask, 'get_event_report')
    def test_scf_unconverged(self, report):
        scf_task = abinit_tasks.ScfFWTask(self.si_scf_input)
        ftm_path = os.path.join(test_dir, "fw_manager_ok.yaml")
        scf_task.ftm = abiflows.fireworks.utils.fw_utils.FWTaskManager.from_file(ftm_path)

        report.return_value = mock_objects.report_ScfConvergenceWarning()

        with mock.patch.object(abinit_tasks.AbiFireTask, 'prepare_restart',
                               return_value=(False, mock_objects.fake_fw, {})) as pr:
            fake_spec = {'test': 1}
            action = scf_task.task_analysis(fake_spec)
            pr.assert_called_once_with(fake_spec)
            self.assertIsInstance(action, FWAction)

            scf_task.restart_info = abinit_tasks.RestartInfo(
                previous_dir='.', num_restarts=scf_task.ftm.fw_policy.max_restarts)
            with self.assertRaises(abinit_tasks.UnconvergedError):
                scf_task.task_analysis(fake_spec)

    @mock.patch.object(abinit_tasks.AbiFireTask, 'get_event_report')
    def test_generic_error(self, report):
        scf_task = abinit_tasks.ScfFWTask(self.si_scf_input)

        report.return_value = mock_objects.report_AbinitError()

        fake_spec = {'test': 1}
        with self.assertRaises(abinit_tasks.AbinitRuntimeError):
            # set the returncode to avoid logging problems
            scf_task.returncode = 10
            scf_task.task_analysis(fake_spec)

    @mock.patch.object(abinit_tasks.AbiFireTask, 'get_event_report')
    def test_no_report_no_err(self, report):
        scf_task = abinit_tasks.ScfFWTask(self.si_scf_input)
        scf_task.set_workdir(os.getcwd())

        report.return_value = None

        fake_spec = {'test': 1}
        with self.assertRaises(abinit_tasks.AbinitRuntimeError):
            # set the returncode to avoid logging problems
            scf_task.returncode = 10
            scf_task.task_analysis(fake_spec)


class TestErrorClasses(AbiflowsTest):

    def setUp(self):
        self.si_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        self.si_scf_input = ebands_input(self.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10).split_datasets()[0]

    @mock.patch.object(abinit_tasks.AbiFireTask, 'get_event_report')
    def test_abinit_runtime_error(self, report):
        err = abinit_tasks.AbinitRuntimeError(msg="test error", num_errors=5)
        err.to_dict()
        new_err = abinit_tasks.AbinitRuntimeError.from_dict(err.as_dict())

        scf_task = abinit_tasks.ScfFWTask(self.si_scf_input)

        report.return_value = mock_objects.report_AbinitError()

        fake_spec = {'test': 1}

        try:
            # set the returncode to avoid logging problems
            scf_task.returncode = 10
            scf_task.task_analysis(fake_spec)
        except abinit_tasks.AbinitRuntimeError as e:
            e.to_dict()
            assert e.num_errors == 1


class TestRestartInfo(AbiflowsTest):

    def test_restart_info(self):
        ri = abinit_tasks.RestartInfo("/path/to/dir", reset=False, num_restarts=4)
        self.assertMSONable(ri)
        ri.prev_indir
        ri.prev_outdir


"""
Tests for the Abinit workflows classes.
The tests require abipy's factories to run, so abipy should be configured and abinit available.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import unittest
import abipy.data as abidata
import abipy.abilab as abilab

from abipy.abio.factories import scf_input
from abiflows.core.testing import AbiflowsTest, has_mongodb, TESTDB_NAME
from abiflows.database.mongoengine.utils import DatabaseData
from abiflows.fireworks.workflows.abinit_workflows import *


class TestBaseClassMethods(AbiflowsTest):

    @classmethod
    def setUpClass(cls):
        cls.si_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        cls.scf_inp = scf_input(cls.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10)
        cls.setup_fireworks()

    @classmethod
    def tearDownClass(cls):
        cls.teardown_fireworks()

    def setUp(self):
        self.scf_wf = ScfFWWorkflow(self.scf_inp)

    def tearDown(self):
        if self.lp:
            self.lp.reset(password=None,require_password=False)

    def test_add_fws(self):
        assert len(self.scf_wf.wf.fws) == 1
        self.scf_wf.add_final_cleanup(out_exts=["WFK", "DEN"])
        assert len(self.scf_wf.wf.fws) == 2
        self.scf_wf.add_mongoengine_db_insertion(DatabaseData(TESTDB_NAME))
        assert len(self.scf_wf.wf.fws) == 3
        self.scf_wf.add_cut3d_den_to_cube_task()
        assert len(self.scf_wf.wf.fws) == 4

    def test_fireworks_methods(self):
        self.scf_wf.add_metadata(self.si_structure, {"test": 1})
        assert "nsites" in self.scf_wf.wf.metadata

        self.scf_wf.fix_fworker("test_worker")
        self.scf_wf.get_reduced_formula(self.scf_inp)
        self.scf_wf.set_short_single_core_to_spec()
        self.scf_wf.set_preserve_fworker()
        self.scf_wf.add_spec_to_all_fws({"test_spec": 1})

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_add_to_db(self):
        self.scf_wf.add_to_db(self.lp)


class TestFromFactory(AbiflowsTest):

    @classmethod
    def setUpClass(cls):
        cls.gan_structure = abilab.Structure.from_file(abidata.cif_file("gan.cif"))
        cls.gan_pseudos = [abidata.pseudos("31ga.pspnc").pseudo_with_symbol('Ga'),
               abidata.pseudos("7n.pspnc").pseudo_with_symbol('N')]

    def test_scf_workflow(self):
        ScfFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                   initialization_info={"test": 1})

        ScfFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                   initialization_info={"test": 1}, autoparal=True)

    def test_phonon_workflow(self):
        PhononFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                      initialization_info={"test": 1}, ph_ngqpt=[2,2,2])

        PhononFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                      initialization_info={"test": 1}, ph_ngqpt=[2,2,2], autoparal=True)

        PhononFullFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                          initialization_info={"test": 1}, qpoints=[[0.1,0,0]])

    def test_dte_workflow(self):
        DteFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, smearing=None,
                                   spin_mode="unpolarized", spec={"test": 1},
                                   initialization_info={"test": 1}, extra_abivars={"ixc": 7})

        DteFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, smearing=None,
                                   spin_mode="unpolarized", spec={"test": 1},
                                   initialization_info={"test": 1}, extra_abivars={"ixc": 7}, autoparal=True)

    def test_relax_workflow(self):
        RelaxFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                     initialization_info={"test": 1})

        RelaxFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                     initialization_info={"test": 1}, autoparal=True)

        RelaxFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                     initialization_info={"test": 1}, target_dilatmx=1.01)

    def test_dfpt_workflow(self):
        # set ixc otherwise the dte part will fail
        extra_abivars = {"ixc": 7}
        DfptFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec={"test": 1},
                                    initialization_info={"test": 1}, ph_ngqpt=[2,2,2], do_ddk=True,
                                    do_dde=True, do_strain=True, do_dte=True, smearing=None,
                                    spin_mode="unpolarized", extra_abivars=extra_abivars)

        DfptFWWorkflow.from_factory(self.gan_structure, self.gan_pseudos, ecut=4, spec=None,
                                    initialization_info=None, ph_ngqpt=[2, 2, 2], do_ddk=True, do_dde=True,
                                    do_strain=False, do_dte=True, extra_abivars=extra_abivars,
                                    smearing=None, spin_mode="unpolarized", autoparal=True)



class TestFromPreviousInput(AbiflowsTest):

    @classmethod
    def setUp(cls):
        cls.gan_structure = abilab.Structure.from_file(abidata.cif_file("gan.cif"))
        cls.gan_pseudos = [abidata.pseudos("31ga.pspnc").pseudo_with_symbol('Ga'),
               abidata.pseudos("7n.pspnc").pseudo_with_symbol('N')]
        cls.scf_inp = scf_input(cls.gan_structure, cls.gan_pseudos, ecut=2, kppa=10, smearing=None,
                                spin_mode="unpolarized")

    def test_phonon_workflow(self):
        PhononFWWorkflow.from_gs_input(gs_input=self.scf_inp, spec={"test": 1},
                                      initialization_info={"test": 1}, ph_ngqpt=[2,2,2])

    def test_dte_workflow(self):
        self.scf_inp['ixc'] = 7
        DteFWWorkflow.from_gs_input(gs_input=self.scf_inp, spec={"test": 1},
                                   initialization_info={"test": 1})

    def test_dfpt_workflow(self):
        self.scf_inp['ixc'] = 7
        DfptFWWorkflow.from_gs_input(gs_input=self.scf_inp, spec=None,
                                     initialization_info=None, ph_ngqpt=[2,2,2],
                                     do_ddk=True,do_dde=True, do_strain=True, do_dte=True)




from __future__ import print_function, division, unicode_literals, absolute_import

import pytest
import os
import glob

from fireworks.core.rocket_launcher import rapidfire
from abipy.electrons.gsr import GsrFile
from abiflows.fireworks.workflows.abinit_workflows import InputFWWorkflow, ScfFWWorkflow
from abiflows.fireworks.tasks.abinit_tasks import ScfFWTask, OUTDIR_NAME, INDIR_NAME, TMPDIR_NAME
from abiflows.fireworks.utils.fw_utils import load_abitask, get_fw_by_task_index
from abiflows.core.testing import AbiflowsIntegrationTest


ABINIT_VERSION = "8.6.1"

# pytestmark = [pytest.mark.skipif(not has_abinit(ABINIT_VERSION), reason="Abinit version {} is not in PATH".format(ABINIT_VERSION)),
#               pytest.mark.skipif(not has_fireworks(), reason="fireworks paackage is missing"),
#               pytest.mark.skipif(not has_mongodb(), reason="no connection to mongodb")]

pytestmark = pytest.mark.usefixtures("cleandb")


class ItestScf(AbiflowsIntegrationTest):

    def itest_input_wf(self, lp, fworker, tmpdir, input_scf_si_low, use_autoparal):
        """
        Tests a simple scf run with the InputFWWorkflow
        """
        wf = InputFWWorkflow(input_scf_si_low, task_type=ScfFWTask, autoparal=use_autoparal)

        scf_fw_id = wf.fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        fw = lp.get_fw_by_id(scf_fw_id)

        assert fw.state == "COMPLETED"

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"
        assert len(wf.leaf_fw_ids) == 1

        if self.check_numerical_values:
            task = load_abitask(lp.get_fw_by_id(wf.leaf_fw_ids[0]))

            with task.open_gsr() as gsr:
                assert gsr.energy == pytest.approx(-241.239839134, rel=0.05)


    def itest_scf_wf(self, lp, fworker, tmpdir, input_scf_si_low, use_autoparal):
        """
        Tests a simple scf run with the ScfFWWorkflow
        """
        wf = ScfFWWorkflow(input_scf_si_low, autoparal=use_autoparal)

        wf.add_final_cleanup(["WFK"])

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        fw = lp.get_fw_by_id(scf_fw_id)

        assert fw.state == "COMPLETED"

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        # check the effect of the final cleanup
        scf_task = load_abitask(get_fw_by_task_index(wf, "scf", index=1))

        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_WFK"))) == 0
        assert len(glob.glob(os.path.join(scf_task.outdir.path, "*_DEN"))) == 1
        assert len(glob.glob(os.path.join(scf_task.tmpdir.path, "*"))) == 0
        assert len(glob.glob(os.path.join(scf_task.indir.path, "*"))) == 0

        if self.check_numerical_values:
            with scf_task.open_gsr() as gsr:
                assert gsr.energy == pytest.approx(-241.239839134, rel=0.05)

    def itest_not_converged(self, lp, fworker, tmpdir, input_scf_si_low):
        """
        Tests the ScfFWWorkflow with a calculation that does not converge on the first run.
        The calculation is continued until convergence is reached
        """

        input_scf_si_low.set_vars(nstep=3)

        wf = ScfFWWorkflow(input_scf_si_low, autoparal=False)

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=1)

        fw = lp.get_fw_by_id(scf_fw_id)

        assert fw.state == "COMPLETED"

        launch = fw.launches[-1]

        assert any(event.yaml_tag == ScfFWTask.CRITICAL_EVENTS[0].yaml_tag
                   for event in launch.action.stored_data['report'])

        links = lp.get_wf_by_fw_id(scf_fw_id).links

        assert scf_fw_id in links and len(links[scf_fw_id]) == 1

        fw_child_id = links[scf_fw_id][0]
        fw_child = lp.get_fw_by_id(fw_child_id)

        assert fw_child.state == "READY"

        # run all the workflow to check that the calculation converged
        rapidfire(lp, fworker, m_dir=str(tmpdir))

        fw_child = lp.get_fw_by_id(fw_child_id)

        assert fw_child.state == "COMPLETED"

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "COMPLETED"

        if self.check_numerical_values:
            task = load_abitask(lp.get_fw_by_id(wf.leaf_fw_ids[0]))

            with task.open_gsr() as gsr:
                assert gsr.energy == pytest.approx(-241.239839133, rel=0.05)


    def itest_not_converged_fizzled(self, lp, fworker, tmpdir, input_scf_si_low):
        """
        Tests the ScfFWWorkflow with a calculation that does not converge within the maximum number
        of restarts allowed. The job will fizzle. The maximum number of restarts is increased and
        the job can continue. This works as a generic test for the maximum number of restarts.
        """

        input_scf_si_low.set_vars(nstep=1)

        # set the maximum number of restarts to 3 through the spec
        # following FWs should have the same option since the spec is passed down when there is a detour
        wf = ScfFWWorkflow(input_scf_si_low, autoparal=False, spec={'fw_policy':{'max_restarts':3}})

        scf_fw_id = wf.scf_fw.fw_id
        old_new = wf.add_to_db(lpad=lp)
        scf_fw_id = old_new[scf_fw_id]

        rapidfire(lp, fworker, m_dir=str(tmpdir))

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert wf.state == "FIZZLED"
        assert len(wf.id_fw) == 4

        # find the fw_id of the last FW
        fw_ids = list(wf.id_fw.keys())
        for father_id, children_ids in wf.links.items():
            if children_ids:
                fw_ids.remove(father_id)
        # there should be only one FW without children
        assert len(fw_ids) == 1
        last_id = fw_ids[0]

        # increase the number of restarts allowed
        lp.update_spec([last_id], {'fw_policy.max_restarts': 10})
        lp.rerun_fw(last_id)

        rapidfire(lp, fworker, m_dir=str(tmpdir), nlaunches=5)

        wf = lp.get_wf_by_fw_id(scf_fw_id)

        assert len(wf.id_fw) == 9
        assert wf.state == 'RUNNING'









# coding: utf-8
"""
Error handlers and validators
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import logging
import os

from abiflows.fireworks.utils.custodian_utils import SRCErrorHandler
from pymatgen.io.abinit.scheduler_error_parsers import MemoryCancelError
from pymatgen.io.abinit.scheduler_error_parsers import MasterProcessMemoryCancelError
from pymatgen.io.abinit.scheduler_error_parsers import SlaveProcessMemoryCancelError
from pymatgen.io.abinit.scheduler_error_parsers import TimeCancelError
from pymatgen.io.abinit.qadapters import QueueAdapter


logger = logging.getLogger(__name__)


class AbinitHandler(SRCErrorHandler):
    """
    General handler for abinit's critical events handlers.
    """

    def __init__(self, job_rundir='.', critical_events=None, queue_adapter=None):
        """
        Initializes the handler with the directory where the job was run.

        Args:
            job_rundir: Directory where the job was run.
        """
        super(AbinitHandler, self).__init__()
        self.job_rundir = job_rundir
        self.critical_events = critical_events

        self.src_fw = False

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'job_rundir': self.job_rundir
                }

    @classmethod
    def from_dict(cls, d):
        return cls(job_rundir=d['job_rundir'])

    @property
    def allow_fizzled(self):
        return False

    @property
    def allow_completed(self):
        return True

    @property
    def handler_priority(self):
        return self.PRIORITY_MEDIUM

    @property
    def skip_remaining_handlers(self):
        return True

    def setup(self):
        if 'SRCScheme' in self.fw_to_check.spec and self.fw_to_check.spec['SRCScheme']:
            self.src_fw = True
        else:
            self.src_fw = False
        self.job_rundir = self.fw_to_check.launches[-1].launch_dir

    def check(self):
        abinit_task = self.fw_to_check.tasks[0]
        self.report = None
        try:
            self.report = abinit_task.get_event_report()
        except Exception as exc:
            msg = "%s exception while parsing event_report:\n%s" % (self, exc)
            logger.critical(msg)

        if self.report is not None:
            # Run has completed, check for critical events (convergence, ...)
            if self.report.run_completed:
                self.events = self.report.filter_types(abinit_task.CRITICAL_EVENTS)
                if self.events:
                    return True
                else:
                    # Calculation has converged
                    # Check if there are custom parameters that should be converged
                    unconverged_params, reset_restart = abinit_task.check_parameters_convergence(self.fw_to_check.spec)
                    if unconverged_params:
                        return True
                    else:
                        return False
            # Abinit run failed to complete
            # Check if the errors can be handled
            if self.report.errors:
                return True
        return True

    def has_corrections(self):
        return True

    def correct(self):
        if self.src_fw:
            if len(self.fw_to_check.tasks) > 1:
                raise ValueError('More than 1 task found in fizzled firework, not yet supported')
            abinit_input_update = {'iscf': 2}
            return {'errors': [self.__class__.__name__],
                    'actions': [{'action_type': 'modify_object',
                                 'object': {'source': 'fw_spec', 'key': 'abinit_input'},
                                 'action': {'_set': abinit_input_update}}]}
        else:
            raise NotImplementedError('This handler cannot be used without the SRC scheme')


class WalltimeHandler(SRCErrorHandler):
    """
    Handler for walltime infringements of the resource manager.
    """

    def __init__(self, job_rundir='.', qout_file='queue.qout', qerr_file='queue.qerr', queue_adapter=None,
                 max_timelimit=None, timelimit_increase=None):
        """
        Initializes the handler with the directory where the job was run, the standard output and error files
        of the queue manager and the queue adapter used.

        Args:
            job_rundir: Directory where the job was run.
            qout_file: Standard output file of the queue manager.
            qerr_file: Standard error file of the queue manager.
            queue_adapter: Queue adapter used to submit the job.
            max_timelimit: Maximum timelimit (in seconds) allowed by the resource manager for the queue.
            timelimit_increase: Amount of time (in seconds) to increase the timelimit.
        """
        super(WalltimeHandler, self).__init__()
        self.job_rundir = job_rundir
        self.qout_file = qout_file
        self.qerr_file = qerr_file
        self.queue_adapter = queue_adapter
        self.setup_filepaths()
        self.max_timelimit = max_timelimit
        self.timelimit_increase = timelimit_increase

        self.src_fw = False

    def setup_filepaths(self):
        self.qout_filepath = os.path.join(self.job_rundir, self.qout_file)
        self.qerr_filepath = os.path.join(self.job_rundir, self.qerr_file)

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'job_rundir': self.job_rundir,
                'qout_file': self.qout_file,
                'qerr_file': self.qerr_file,
                'queue_adapter': self.queue_adapter.as_dict() if self.queue_adapter is not None else None,
                'max_timelimit': self.max_timelimit,
                'timelimit_increase': self.timelimit_increase
                }

    @classmethod
    def from_dict(cls, d):
        qa = QueueAdapter.from_dict(d['queue_adapter']) if d['queue_adapter'] is not None else None
        return cls(job_rundir=d['job_rundir'], qout_file=d['qout_file'], qerr_file=d['qerr_file'], queue_adapter=qa,
                   max_timelimit=d['max_timelimit'],
                   timelimit_increase=d['timelimit_increase'])

    @property
    def allow_fizzled(self):
        return True

    @property
    def allow_completed(self):
        return False

    @property
    def handler_priority(self):
        return self.PRIORITY_VERY_LOW

    @property
    def skip_remaining_handlers(self):
        return True

    def setup(self):
        if 'SRCScheme' in self.fw_to_check.spec and self.fw_to_check.spec['SRCScheme']:
            self.src_fw = True
        else:
            self.src_fw = False
        self.job_rundir = self.fw_to_check.launches[-1].launch_dir
        self.setup_filepaths()
        self.queue_adapter = self.fw_to_check.spec['qtk_queueadapter']

    def check(self):

        # Analyze the stderr and stdout files of the resource manager system.
        qerr_info = None
        qout_info = None
        if os.path.exists(self.qerr_filepath):
            with open(self.qerr_filepath, "r") as f:
                qerr_info = f.read()
        if os.path.exists(self.qout_filepath):
            with open(self.qout_filepath, "r") as f:
                qout_info = f.read()

        self.timelimit_error = None
        self.queue_errors = None
        if qerr_info or qout_info:
            from pymatgen.io.abinit.scheduler_error_parsers import get_parser
            qtype = self.queue_adapter.QTYPE
            scheduler_parser = get_parser(qtype, err_file=self.qerr_filepath,
                                          out_file=self.qout_filepath)

            if scheduler_parser is None:
                raise ValueError('Cannot find scheduler_parser for qtype {}'.format(qtype))

            scheduler_parser.parse()
            self.queue_errors = scheduler_parser.errors

            for error in self.queue_errors:
                if isinstance(error, TimeCancelError):
                    logger.debug('found timelimit error.')
                    self.timelimit_error = error
                    return True
        return False

    def correct(self):
        if self.src_fw:
            if len(self.fw_to_check.tasks) > 1:
                raise ValueError('More than 1 task found in "memory-fizzled" firework, not yet supported')
            logger.debug('adding SRC detour')
            # Information about the update of the memory (master overhead or base mem per proc) in the queue adapter
            queue_adapter_update = {}
            # When max_timelimit is not set, automatically take the hard timelimit of the queue
            if self.max_timelimit is None:
                max_timelimit = self.queue_adapter.timelimit_hard
            else:
                max_timelimit = self.max_timelimit
            # When timelimit_increase is not set, automatically take a tenth of the hard timelimit of the queue
            if self.timelimit_increase is None:
                timelimit_increase = self.queue_adapter.timelimit_hard / 10
            else:
                timelimit_increase = self.timelimit_increase
            if isinstance(self.timelimit_error, TimeCancelError):
                old_timelimit = self.queue_adapter.timelimit
                if old_timelimit == max_timelimit:
                    raise ValueError('Cannot increase beyond maximum timelimit ({:d} seconds) set in WalltimeHandler.'
                                     'Hard time limit of '
                                     'the queue is {:d} seconds'.format(max_timelimit,
                                                                        self.queue_adapter.timelimit_hard))
                new_timelimit = old_timelimit + timelimit_increase
                # If the new timelimit exceeds the max timelimit, just put it to the max timelimit
                if new_timelimit > max_timelimit:
                    new_timelimit = max_timelimit
                queue_adapter_update['timelimit'] = new_timelimit
            else:
                raise ValueError('Should not be here ...')
            return {'errors': [self.__class__.__name__],
                    'actions': [{'action_type': 'modify_object',
                                 'object': {'source': 'fw_spec', 'key': 'qtk_queueadapter'},
                                 'action': {'_set': queue_adapter_update}}]}
        else:
            raise NotImplementedError('This handler cannot be used without the SRC scheme')

    def has_corrections(self):
        return True


class MemoryHandler(SRCErrorHandler):
    """
    Handler for memory infringements of the resource manager. The handler should be able to handle the possible
    overhead of the master process.
    """

    def __init__(self, job_rundir='.', qout_file='queue.qout', qerr_file='queue.qerr', queue_adapter=None,
                 max_mem_per_proc_mb=8000, mem_per_proc_increase_mb=1000,
                 max_master_mem_overhead_mb=8000, master_mem_overhead_increase_mb=1000):
        """
        Initializes the handler with the directory where the job was run, the standard output and error files
        of the queue manager and the queue adapter used.

        Args:
            job_rundir: Directory where the job was run.
            qout_file: Standard output file of the queue manager.
            qerr_file: Standard error file of the queue manager.
            queue_adapter: Queue adapter used to submit the job.
            max_mem_per_proc_mb: Maximum memory per process in megabytes.
            mem_per_proc_increase_mb: Amount of memory to increase the memory per process in megabytes.
            max_master_mem_overhead_mb: Maximum overhead memory for the master process in megabytes.
            master_mem_overhead_increase_mb: Amount of memory to increase the overhead memory for the master process
                                             in megabytes.
        """
        super(MemoryHandler, self).__init__()
        self.job_rundir = job_rundir
        self.qout_file = qout_file
        self.qerr_file = qerr_file
        self.queue_adapter = queue_adapter
        self.setup_filepaths()
        self.max_mem_per_proc_mb = max_mem_per_proc_mb
        self.mem_per_proc_increase_mb = mem_per_proc_increase_mb
        self.max_master_mem_overhead_mb = max_master_mem_overhead_mb
        self.master_mem_overhead_increase_mb = master_mem_overhead_increase_mb

        self.src_fw = False

    def setup_filepaths(self):
        self.qout_filepath = os.path.join(self.job_rundir, self.qout_file)
        self.qerr_filepath = os.path.join(self.job_rundir, self.qerr_file)

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'job_rundir': self.job_rundir,
                'qout_file': self.qout_file,
                'qerr_file': self.qerr_file,
                'queue_adapter': self.queue_adapter.as_dict() if self.queue_adapter is not None else None,
                'max_mem_per_proc_mb': self.max_mem_per_proc_mb,
                'mem_per_proc_increase_mb': self.mem_per_proc_increase_mb,
                'max_master_mem_overhead_mb': self.max_master_mem_overhead_mb,
                'master_mem_overhead_increase_mb': self.master_mem_overhead_increase_mb
                }

    @classmethod
    def from_dict(cls, d):
        qa = QueueAdapter.from_dict(d['queue_adapter']) if d['queue_adapter'] is not None else None
        return cls(job_rundir=d['job_rundir'], qout_file=d['qout_file'], qerr_file=d['qerr_file'], queue_adapter=qa,
                   max_mem_per_proc_mb=d['max_mem_per_proc_mb'],
                   mem_per_proc_increase_mb=d['mem_per_proc_increase_mb'],
                   max_master_mem_overhead_mb=d['max_master_mem_overhead_mb'],
                   master_mem_overhead_increase_mb=d['master_mem_overhead_increase_mb'])

    @property
    def allow_fizzled(self):
        return True

    @property
    def allow_completed(self):
        return False

    @property
    def handler_priority(self):
        return self.PRIORITY_VERY_LOW

    @property
    def skip_remaining_handlers(self):
        return True

    def setup(self):
        if 'SRCScheme' in self.fw_to_check.spec and self.fw_to_check.spec['SRCScheme']:
            self.src_fw = True
        else:
            self.src_fw = False
        self.job_rundir = self.fw_to_check.launches[-1].launch_dir
        self.setup_filepaths()
        self.queue_adapter = self.fw_to_check.spec['qtk_queueadapter']

    def check(self):

        # Analyze the stderr and stdout files of the resource manager system.
        qerr_info = None
        qout_info = None
        if os.path.exists(self.qerr_filepath):
            with open(self.qerr_filepath, "r") as f:
                qerr_info = f.read()
        if os.path.exists(self.qout_filepath):
            with open(self.qout_filepath, "r") as f:
                qout_info = f.read()

        self.memory_error = None
        self.queue_errors = None
        if qerr_info or qout_info:
            from pymatgen.io.abinit.scheduler_error_parsers import get_parser
            qtype = self.queue_adapter.QTYPE
            scheduler_parser = get_parser(qtype, err_file=self.qerr_filepath,
                                          out_file=self.qout_filepath)

            if scheduler_parser is None:
                raise ValueError('Cannot find scheduler_parser for qtype {}'.format(qtype))

            scheduler_parser.parse()
            self.queue_errors = scheduler_parser.errors

            #TODO: handle the cases where it is Master or Slave here ... ?
            for error in self.queue_errors:
                if isinstance(error, MemoryCancelError):
                    logger.debug('found memory error.')
                    self.memory_error = error
                    return True
                if isinstance(error, MasterProcessMemoryCancelError):
                    logger.debug('found master memory error.')
                    self.memory_error = error
                    return True
                if isinstance(error, SlaveProcessMemoryCancelError):
                    logger.debug('found slave memory error.')
                    self.memory_error = error
                    return True
        return False

    def correct(self):
        if self.src_fw:
            if len(self.fw_to_check.tasks) > 1:
                raise ValueError('More than 1 task found in "memory-fizzled" firework, not yet supported')
            logger.debug('adding SRC detour')
            # Information about the update of the memory (master overhead or base mem per proc) in the queue adapter
            queue_adapter_update = {}
            if isinstance(self.memory_error, (MemoryCancelError, SlaveProcessMemoryCancelError)):
                old_mem_per_proc = self.queue_adapter.mem_per_proc
                new_mem_per_proc = old_mem_per_proc + self.mem_per_proc_increase_mb
                queue_adapter_update['mem_per_proc'] = new_mem_per_proc
            elif isinstance(self.memory_error, MasterProcessMemoryCancelError):
                old_mem_overhead = self.queue_adapter.master_mem_overhead
                new_mem_overhead = old_mem_overhead + self.master_mem_overhead_increase_mb
                if new_mem_overhead > self.max_master_mem_overhead_mb:
                    raise ValueError('New master memory overhead {:d} is larger than '
                                     'max master memory overhead {:d}'.format(new_mem_overhead,
                                                                              self.max_master_mem_overhead_mb))
                queue_adapter_update['master_mem_overhead'] = new_mem_overhead
            else:
                raise ValueError('Should not be here ...')
            return {'errors': [self.__class__.__name__],
                    'actions': [{'action_type': 'modify_object',
                                 'object': {'source': 'fw_spec', 'key': 'qtk_queueadapter'},
                                 'action': {'_set': queue_adapter_update}}]}
        else:
            raise NotImplementedError('This handler cannot be used without the SRC scheme')

    def has_corrections(self):
        return True


class UltimateMemoryHandler(MemoryHandler):
    """
    Handler for infringements of the resource manager. If no memory error is found,
    """

    def __init__(self, job_rundir='.', qout_file='queue.qout', qerr_file='queue.qerr', queue_adapter=None,
                 max_mem_per_proc_mb=8000, mem_per_proc_increase_mb=1000,
                 max_master_mem_overhead_mb=8000, master_mem_overhead_increase_mb=1000):
        """
        Initializes the handler with the directory where the job was run, the standard output and error files
        of the queue manager and the queue adapter used.

        Args:
            job_rundir: Directory where the job was run.
            qout_file: Standard output file of the queue manager.
            qerr_file: Standard error file of the queue manager.
            queue_adapter: Queue adapter used to submit the job.
            max_mem_per_proc_mb: Maximum memory per process in megabytes.
            mem_per_proc_increase_mb: Amount of memory to increase the memory per process in megabytes.
            max_master_mem_overhead_mb: Maximum overhead memory for the master process in megabytes.
            master_mem_overhead_increase_mb: Amount of memory to increase the overhead memory for the master process
                                             in megabytes.
        """
        super(UltimateMemoryHandler, self).__init__()

    @property
    def handler_priority(self):
        return self.PRIORITY_LAST

    def check(self):
        mem_check = super(UltimateMemoryHandler, self).check()
        if mem_check:
            raise ValueError('This error should have been caught by a standard MemoryHandler ...')
        #TODO: Do we have some check that we can do here ?
        return True

    def correct(self):
        if self.src_fw:
            if len(self.fw_to_check.tasks) > 1:
                raise ValueError('More than 1 task found in "memory-fizzled" firework, not yet supported')
            if self.memory_error is not None:
                raise ValueError('This error should have been caught by a standard MemoryHandler ...')
            if self.queue_errors is not None and len(self.queue_errors) > 0:
                raise ValueError('Queue errors were found ... these should be handled properly by another handler')
            # Information about the update of the memory (base mem per proc) in the queue adapter
            queue_adapter_update = {}
            old_mem_per_proc = self.queue_adapter.mem_per_proc
            new_mem_per_proc = old_mem_per_proc + self.mem_per_proc_increase_mb
            queue_adapter_update['mem_per_proc'] = new_mem_per_proc
            return {'errors': [self.__class__.__name__],
                    'actions': [{'action_type': 'modify_object',
                                 'object': {'source': 'fw_spec', 'key': 'qtk_queueadapter'},
                                 'action': {'_set': queue_adapter_update}}]}
        else:
            raise NotImplementedError('This handler cannot be used without the SRC scheme')


# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

from abiflows.core.testing import AbiflowsTest
from abiflows.fireworks.tasks.src_tasks_abc import SRCCleanerOptions


class TestSRCCleanerOptions(AbiflowsTest):
    def test_steps_to_clean(self):
        # Clean all option
        opt = SRCCleanerOptions.clean_all()
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2, 3, 4])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Clean all except last option
        opt = SRCCleanerOptions.clean_all_except_last()
        to_clean = opt.steps_to_clean(this_step_index=6, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2, 3, 4, 5])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])

        # Test errors raised
        self.assertRaises(ValueError, SRCCleanerOptions, when_to_clean='EACH_STEP',
                          current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                          which_src_steps_to_clean='the_one_before_the_7_previous_one')
                                        #should be "the_one_before_the_7_previous_ones"
        self.assertRaises(ValueError, SRCCleanerOptions, when_to_clean='EACH_STEP',
                          current_src_states_allowed=['RECOVERABLES', 'FINALIZED'],
                          which_src_steps_to_clean='the_one_before_the_7_previous_ones')
                                        #"RECOVERABLES" should be "RECOVERABLE"
        self.assertRaises(ValueError, SRCCleanerOptions, when_to_clean='EACH_STEP',
                          current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                          which_src_steps_to_clean='single_0')
                                        #a single step index to be cleaned should be >= 1
        self.assertRaises(ValueError, SRCCleanerOptions, when_to_clean='ALL_STEPS',
                          current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                          which_src_steps_to_clean='single_0')
                                    #when_to_clean should be one of 'EACH_STEP', 'LAST_STEP', 'EACH_STEP_EXCEPT_LAST'

        # Testing all
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                                which_src_steps_to_clean='all')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2, 3, 4])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Testing this_one
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['FINALIZED'],
                                which_src_steps_to_clean='this_one')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [4])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='RECOVERABLE')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Testing all_before_this_one
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['FINALIZED'],
                                which_src_steps_to_clean='all_before_this_one')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2, 3])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='RECOVERABLE')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=2, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Testing all_before_the_previous_one
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['FINALIZED'],
                                which_src_steps_to_clean='all_before_the_previous_one')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='RECOVERABLE')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=2, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=3, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Testing the_one_before_this_one
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                                which_src_steps_to_clean='the_one_before_this_one')
        to_clean = opt.steps_to_clean(this_step_index=6, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [5])
        to_clean = opt.steps_to_clean(this_step_index=2, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])

        # Testing the_one_before_the_previous_one
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['RECOVERABLE', 'FINALIZED'],
                                which_src_steps_to_clean='the_one_before_the_previous_one')
        to_clean = opt.steps_to_clean(this_step_index=6, this_step_state='UNRECOVERABLE')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=6, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [4])
        to_clean = opt.steps_to_clean(this_step_index=2, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])

        # Testing all_before_the_N_previous_ones
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['FINALIZED'],
                                which_src_steps_to_clean='all_before_the_3_previous_ones')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=3, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=5, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])
        to_clean = opt.steps_to_clean(this_step_index=8, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1, 2, 3, 4])
        to_clean = opt.steps_to_clean(this_step_index=8, this_step_state='RECOVERABLE')
        self.assertEqual(to_clean, [])

        # Testing the_one_before_the_N_previous_ones
        opt = SRCCleanerOptions(when_to_clean='EACH_STEP',
                                current_src_states_allowed=['FINALIZED'],
                                which_src_steps_to_clean='the_one_before_the_2_previous_ones')
        to_clean = opt.steps_to_clean(this_step_index=4, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [1])
        to_clean = opt.steps_to_clean(this_step_index=1, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=3, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [])
        to_clean = opt.steps_to_clean(this_step_index=5, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [2])
        to_clean = opt.steps_to_clean(this_step_index=8, this_step_state='FINALIZED')
        self.assertEqual(to_clean, [5])
        to_clean = opt.steps_to_clean(this_step_index=8, this_step_state='RECOVERABLE')
        self.assertEqual(to_clean, [])



        # ['all', 'this_one', 'all_before_this_one', 'all_before_the_previous_one',
        #                                 'the_one_before_this_one', 'the_one_before_the_previous_one']


# coding: utf-8
"""
Abstract base classes for controllers, monitors, and other possible tools/utils/objects used to manage, correct,
monitor and check tasks, events, results, objects, ...
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import abc
import fnmatch
import logging
import os
import shutil
import traceback

from six import add_metaclass
from monty.json import MontyDecoder
from monty.json import MSONable


logger = logging.getLogger(__name__)


PRIORITIES = {'PRIORITY_HIGHEST': 1000,
              'PRIORITY_VERY_HIGH': 875,
              'PRIORITY_HIGH': 750,
              'PRIORITY_MEDIUM_HIGH': 625,
              'PRIORITY_MEDIUM': 500,
              'PRIORITY_MEDIUM_LOW': 375,
              'PRIORITY_LOW': 250,
              'PRIORITY_VERY_LOW': 125,
              'PRIORITY_LOWEST': 0}

PRIORITY_HIGHEST = PRIORITIES['PRIORITY_HIGHEST']
PRIORITY_VERY_HIGH = PRIORITIES['PRIORITY_VERY_HIGH']
PRIORITY_HIGH = PRIORITIES['PRIORITY_HIGH']
PRIORITY_MEDIUM_HIGH = PRIORITIES['PRIORITY_MEDIUM_HIGH']
PRIORITY_MEDIUM = PRIORITIES['PRIORITY_MEDIUM']
PRIORITY_MEDIUM_LOW = PRIORITIES['PRIORITY_MEDIUM_LOW']
PRIORITY_LOW = PRIORITIES['PRIORITY_LOW']
PRIORITY_VERY_LOW = PRIORITIES['PRIORITY_VERY_LOW']
PRIORITY_LOWEST = PRIORITIES['PRIORITY_LOWEST']


#TODO: find a good name (is barrier ok ?). This is a container of controllers ...
# class ControlStep(MSONable):
# class ControlStage(MSONable):
# class ControlStation(MSONable):
# class ControlGate(MSONable):
# class ControlProcess(MSONable):
# class ControlBarrier(MSONable):
class ControlProcedure(MSONable):

    def __init__(self, controllers, monitors=None, sorting=None):
        self.controllers = []
        self.add_controllers(controllers=controllers)
        self.controlled_item_type = None

    def set_controlled_item_type(self, controlled_item_type):
        self.controlled_item_type = controlled_item_type

    def setup_controllers(self, controlled_item_type):
        self.grouped_controllers = {}
        controllers = [controller for controller in self.controllers
                       if controlled_item_type in controller.controlled_item_types]
        for controller in controllers:
            if controller.priority in self.grouped_controllers:
                self.grouped_controllers[controller.priority].append(controller)
            else:
                self.grouped_controllers[controller.priority] = [controller]
        self.priorities = sorted(self.grouped_controllers.keys(), reverse=True)
        self._ncontrollers = sum([len(v) for v in self.grouped_controllers.values()])

    def add_controller(self, controller):
        self.add_controllers(controllers=[controller])

    def add_controllers(self, controllers):
        if isinstance(controllers, (list, tuple)):
            for controller in controllers:
                if not issubclass(controller.__class__, Controller):
                    raise ValueError('One of the controllers is not a subclass of Controller')
            self.controllers.extend(controllers)
        elif issubclass(controllers.__class__, Controller):
            self.controllers.append(controllers)
        else:
            raise ValueError('controllers should be either a list of subclasses of Controller or a single '
                             'subclass of Controller')

    def process(self, **kwargs):
        self.setup_controllers(controlled_item_type=self.controlled_item_type)
        report = ControlReport()
        if self.ncontrollers == 0:
            report.state = ControlReport.UNRECOVERABLE
            return report
        for priority in self.priorities:
            skip_lower_priority = False
            #skip_other_controllers = False
            for controller in self.grouped_controllers[priority]:
                if controller._only_unfinalized and report.finalized:
                    #TODO: clean up here ... otherwise the ultimate could always be applied .....
                    continue
                controller_note = controller.process(**kwargs)
                # TODO: clean up this ...
                report.add_controller_note(controller_note=controller_note)
            #    if controller_note.state in ControllerNote.ERROR_STATES:
            #        skip_other_controllers = True
            #        break
                if report.state == ControlReport.RECOVERABLE and controller.skip_lower_priority_controllers:
                    skip_lower_priority = True
            if skip_lower_priority:
                break
            #if skip_other_controllers:
            #    break
        # if report.finalized and self.cleaner is not None:
        #     self.cleaner.clean()
        return report

    @property
    def ncontrollers(self):
        return self._ncontrollers

    @classmethod
    def from_dict(cls, d):
        dec = MontyDecoder()
        return cls(controllers=dec.process_decoded(d['controllers']))

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'controllers': [controller.as_dict() for controller in self.controllers]}


class ControlledItemType(MSONable):
    ALLOWED_CONTROL_ITEMS = {'TASK': ['READY', 'NOT_READY', 'VALID'],
                             'TASK_BEFORE_EXECUTION': ['READY', 'NOT_READY'],
                             'TASK_RUNNING': ['VALID', 'RECOVERABLE'],
                             'TASK_ABORTED': ['RECOVERABLE', 'UNRECOVERABLE'],
                             'TASK_FAILED': ['RECOVERABLE', 'UNRECOVERABLE', 'UNKNOWN_REASON'],
                             'TASK_COMPLETED': ['VALID', 'RECOVERABLE', 'UNRECOVERABLE'],
                             'FILE': ['MISSING', 'EMPTY', 'CORRUPTED', 'ERRONEOUS', 'VALID'],
                             'OBJECT': ['CORRUPTED', 'ERRONEOUS', 'VALID']}

    #     # Status of a controlled failed task
    # TASK_FAILED_UNRECOVERABLE_STATUS = 'TASK_FAILED_UNRECOVERABLE'
    # TASK_FAILED_UNKNOWN_REASON_STATUS = 'TASK_FAILED_UNKNOWN_REASON'
    # TASK_FAILED_RECOVERABLE_STATUS = 'TASK_FAILED_RECOVERABLE'
    #
    # # Status of a controlled stopped task
    # TASK_STOPPED_UNRECOVERABLE_STATUS = 'TASK_STOPPED_UNRECOVERABLE'
    # TASK_STOPPED_RECOVERABLE_STATUS = 'TASK_STOPPED_RECOVERABLE'
    #
    # # Status of a controlled completed task
    # TASK_COMPLETED_UNRECOVERABLE_STATUS = 'TASK_COMPLETED_UNRECOVERABLE'
    # TASK_COMPLETED_RECOVERABLE_STATUS = 'TASK_COMPLETED_RECOVERABLE'
    #
    # # Status of a controlled file
    # FILE_MISSING_STATUS = 'FILE_MISSING'
    # FILE_EMPTY_STATUS = 'FILE_EMPTY'
    # FILE_CORRUPTED_STATUS = 'FILE_CORRUPTED'
    # FILE_ERRONEOUS_STATUS = 'FILE_ERRONEOUS'
    # FILE_CORRECT_STATUS = 'FILE_CORRECT'
    #
    # # Status of a controlled object
    # OBJECT_CORRUPTED_STATUS = 'OBJECT_CORRUPTED'
    # OBJECT_ERRONEOUS_STATUS = 'OBJECT_ERRONEOUS'
    # OBJECT_CORRECT_STATUS = 'OBJECT_CORRECT'

    def __init__(self, item_type):
        self._set_item(item_type=item_type)

    def _set_item(self, item_type):
        if item_type not in self.ALLOWED_CONTROL_ITEMS:
            raise ValueError('"item_type" should be one of the following :'
                             ' {}'.format(', '.join(['"{}"' for ii in self.ALLOWED_CONTROL_ITEMS])))
        self._item_type = item_type

    @classmethod
    def task(cls):
        return cls(item_type='TASK')

    @classmethod
    def task_running(cls):
        return cls(item_type='TASK_RUNNING')

    @classmethod
    def task_aborted(cls):
        return cls(item_type='TASK_ABORTED')

    @classmethod
    def task_failed(cls):
        return cls(item_type='TASK_FAILED')

    @classmethod
    def task_completed(cls):
        return cls(item_type='TASK_COMPLETED')

    @classmethod
    def file(cls):
        return cls(item_type='FILE')

    @classmethod
    def object(cls):
        return cls(item_type='OBJECT')

    def __eq__(self, other):
        return self._item_type == other._item_type

    def __hash__(self):
        return self._item_type

    @classmethod
    def from_dict(cls, d):
        return cls(item_type=d['item_type'])

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'item_type': self._item_type}



@add_metaclass(abc.ABCMeta)
class Monitor:
    pass

@add_metaclass(abc.ABCMeta)
#class ControlStep(MSONable):
class Controller(MSONable):
    """
    Abstract base class for controlling a task, an event, a result, an output, an object, ...
    """

    _priority = PRIORITY_MEDIUM
    _controlled_item_types = None
    #TODO: try to make this cleaner ...
    _only_unfinalized = False

    # Types of controllers
    # Combinations of the following are possible, e.g. some controller might be a monitor, a handler, a manager and a
    #  validator at the same time

    # 2. Handler : Whether this controller is supposed to handle errors
    #              States that handlers can specify in their note : "NOTHING_FOUND", "ERROR_UNRECOVERABLE",
    #                                                               "ERROR_NOFIX", "ERROR_FIXSTOP", "ERROR_FIXCONTINUE"
    # is_handler = False

    # 4. Validator : Whether this controller is used to validate the results/tasks/...
    #                States that validators can specify in their note : "OK", "NOT_OK"
    # is_validator = False
    # NB: - The distinction between handler and manager is thin and is actually more up to the user. At this moment,
    #        they both are at the exact same level in the implementation ...

    can_validate = False

    def __init__(self):
        pass

    @abc.abstractmethod
    def from_dict(cls, d):
        pass

    @abc.abstractmethod
    def as_dict(self):
        pass

    @abc.abstractmethod
    def process(self, **kwargs):
        """
        Main function used to make the actual control/check of a list of inputs/outputs.
        The function should return a ControllerNote object containing the main conclusion of the controller, i.e.
        whether something important has been detected, as well as the possible actions/corrections to be done
        in order to carry on/continue/restart a task.
        """
        pass

    def set_priority(self, priority):
        if priority in PRIORITIES.keys():
            self._priority = PRIORITIES[priority]
        elif isinstance(priority, int) and 0 <= priority <= 1000:
            self._priority = priority
        else:
            raise ValueError('"priority" in set_priority should be either an integer between 0 and 1000 or '
                             'one of the following : {}'.format(', '.join([str(ii) for ii in PRIORITIES.keys()])))

    @property
    def priority(self):
        return self._priority

    @priority.setter
    def priority(self, priority):
        if priority in PRIORITIES.keys():
            self._priority = PRIORITIES[priority]
        elif isinstance(priority, int) and 0 <= priority <= 1000:
            self._priority = priority
        else:
            raise ValueError('"priority" should be either an integer between 0 and 1000 or '
                             'one of the following : {}'.format(', '.join([str(ii) for ii in PRIORITIES.keys()])))

    @property
    def controlled_item_types(self):
        return self._controlled_item_types

    @property
    def skip_remaining_controllers(self):
        return False

    @property
    def skip_lower_priority_controllers(self):
        return False

    @property
    def validated(self):
        return None


class Cleaner(MSONable):
    def __init__(self, dirs_and_patterns):
        """
        Initializes a Cleaner object used to delete files. The cleaner takes a simple list of dict described below
        :param dirs_and_patterns: list of dicts with keys "directory" and "patterns". for each item in the list, any
                                  file or directory present in the "directory" and matching one of the "patterns" will
                                  be deleted.
        Example : dirs_and_patterns = [{'directory': 'out',
                                        'patterns': ['*.log', '*.backup']},
                                       {'directory': 'tmp',
                                        'patterns': ['*']}]
                  will remove all files ending with ".log" or ".backup" in the "out" directory and all files in the
                  "tmp" directory.

        """
        for dir_and_patterns in dirs_and_patterns:
            if os.path.isabs(dir_and_patterns['directory']):
                raise ValueError('The definition of the directories to clean should be as relative paths and not '
                                 'absolute paths')
        self.dirs_and_patterns = dirs_and_patterns

    def clean(self, root_directory):
        if not os.path.isabs(root_directory):
            raise ValueError('The root directory to clean should be defined with an absolute path')
        deleted_files = []
        for dir_and_patterns in self.dirs_and_patterns:
            directory = os.path.join(root_directory, dir_and_patterns['directory'])
            if os.path.isdir(directory):
                for file in os.listdir(directory):
                    for patt in dir_and_patterns['patterns']:
                        if fnmatch.fnmatch(file, patt):
                            fp = os.path.join(directory, file)
                            try:
                                if os.path.isfile(fp):
                                    os.unlink(fp)
                                elif os.path.isdir(fp):
                                    shutil.rmtree(fp)
                                deleted_files.append(fp)
                                break
                            except:
                                logger.warning("Couldn't delete {}: {}".format(fp, traceback.format_exc()))
        pass

    # def delete_files(d, exts=None):
    #     deleted_files = []
    #     if os.path.isdir(d):
    #         for f in os.listdir(d):
    #             if exts is None or "*" in exts or any(ext in f for ext in exts):
    #                 fp = os.path.join(d, f)
    #                 try:
    #                     if os.path.isfile(fp):
    #                         os.unlink(fp)
    #                     elif os.path.isdir(fp):
    #                         shutil.rmtree(fp)
    #                     deleted_files.append(fp)
    #                 except:
    #                     logger.warning("Couldn't delete {}: {}".format(fp, traceback.format_exc()))
    #
    #     return deleted_files


@add_metaclass(abc.ABCMeta)
class Manager(MSONable):
    def __init__(self):
        pass

    @abc.abstractmethod
    def manage(self):
        pass

# class ControllerStatement(MSONable):
# class ControllerAccount(MSONable):
# class ControllerRecord(MSONable):
class ControllerNote(MSONable):

    # Special state returned by a controller that is supposed to be "more important" than others and can say whether
    #  a given task is completely finalized and does not need further restarts/changes/... In that case, it is very
    #  clear that the task is completed.
    EVERYTHING_OK = 'EVERYTHING_OK'
    # State of a controlled task specifying that nothing was detected by the controller
    NOTHING_FOUND = 'NOTHING_FOUND'
    # State of a controlled task specifying that some error(s) was (were) detected by the controller and this (these)
    #  error(s) is unrecoverable. In that case, it is very clear that nothing can be done.
    ERROR_UNRECOVERABLE = 'ERROR_UNRECOVERABLE'
    # State of a controlled task specifying that some error(s) was (were) detected by the controller but the controller
    #  could not fix that error. In that case, it is possible that some other controller could fix the situation
    ERROR_NOFIX = 'ERROR_NOFIX'
    # State of a controlled task specifying that some error(s) was (were) detected by the controller and the controller
    #  adviced some action to fix the error. No other controller should be applied.
    ERROR_RECOVERABLE = 'ERROR_RECOVERABLE'
    # # State of a controlled task specifying that some error(s) was (were) detected by the controller and the controller
    # #  adviced some action to fix the error. Other controllers (if any) might still be applied.
    # ERROR_FIXCONTINUE = 'ERROR_FIXCONTINUE'

    STATES = [EVERYTHING_OK, NOTHING_FOUND, ERROR_UNRECOVERABLE, ERROR_NOFIX, ERROR_RECOVERABLE]
    ERROR_STATES = [ERROR_UNRECOVERABLE, ERROR_NOFIX, ERROR_RECOVERABLE]

    #TODO consider using increasing integers as values, so that we can take the lowest as a general value of the
    # restart
    # Restart from the very beginning of the calculation. Every intermediate result should be neglected
    RESTART_FROM_SCRATCH = "RESTART_FROM_SCRATCH"

    # Restart from the current step but don't make use previous informations
    RESET_RESTART = "RESET_RESTART"

    # Restart with the maximum amount of information available
    SIMPLE_RESTART = "SIMPLE_RESTART"

    def __init__(self, controller, state=None, problems=None, actions=None, restart=None, is_valid=None):
        self.controller = controller
        self.state = state
        self.problems = problems
        self.actions = actions
        self.restart = restart
        self.is_valid = is_valid

    @property
    def actions(self):
        return self._actions

    @actions.setter
    def actions(self, actions):
        if actions is None:
            actions = {}
        self._actions = actions

    def add_problem(self, problem):
        if self.problems is None:
            self.problems = [problem]
        else:
            self.problems.append(problem)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if state is None:
            self._state = None
        elif state in self.STATES:
            self._state = state
        else:
            raise ValueError('"state" in ControllerNote should be one of the following : '
                             '{}'.format(', '.join(self.STATES)))

    @property
    def has_errors_recoverable(self):
        return True

    @property
    def has_errors_unrecoverable(self):
        return True

    @property
    def is_recoverable(self):
        return True

    @property
    def has_errors(self):
        return True

    @property
    def is_valid(self):
        if not self.controller.can_validate:
            raise ValueError('This ControllerNote comes from a controller that cannot validate !')
        return self._is_valid

    @is_valid.setter
    def is_valid(self, is_valid):
        self._is_valid = is_valid

    def set_restart(self, restart):
        self.restart = restart

    def simple_restart(self):
        self.restart = self.SIMPLE_RESTART

    def reset_restart(self):
        self.restart = self.RESET_RESTART

    def restart_from_scratch(self):
        self.restart = self.RESTART_FROM_SCRATCH

    @classmethod
    def from_dict(cls, d):
        dec = MontyDecoder()
        if 'is_valid' in d:
            return cls(controller=dec.process_decoded(d['controller']),
                       state=d['state'], problems=d['problems'],
                       actions=dec.process_decoded(d['actions']),
                       restart=d['restart'], is_valid=d['is_valid'])
        else:
            return cls(controller=dec.process_decoded(d['controller']),
                       state=d['state'], problems=d['problems'],
                       actions=dec.process_decoded(d['actions']),
                       restart=d['restart'])

    def as_dict(self):
        dd = {'@class': self.__class__.__name__,
              '@module': self.__class__.__module__,
              'controller': self.controller.as_dict(),
              'state': self.state,
              'problems': self.problems,
              'actions': {key: action.as_dict() for key, action in self.actions.items()},
              'restart': self.restart}
        if self.controller.can_validate:
            dd['is_valid'] = self.is_valid
        return dd


class ControlReport(MSONable):

    # Status of an SRC trio that is completely finalized
    FINALIZED = 'FINALIZED'
    # Status of an SRC trio that has one or more errors that is (are) unrecoverable
    UNRECOVERABLE = 'UNRECOVERABLE'
    # Status of an SRC trio that has one or more errors that is (are) recoverable
    RECOVERABLE = 'RECOVERABLE'
    # Status of an SRC trio that is ongoing (for controllers taking care of non-error-related stuff, e.g. convergence,
    #  accuracy goal achieved with multiple-steps such as relaxation with low ecut then high ecut, ...)
    ONGOING = 'ONGOING'
    NONE = 'NONE'

    STATES = [FINALIZED, UNRECOVERABLE, RECOVERABLE, ONGOING, NONE]

    def __init__(self, controller_notes=None):
        self.controller_notes = []
        if controller_notes is not None:
            self.add_controller_notes(controller_notes)
        self.update_state_from_controller_notes()

    def add_controller_notes(self, controller_notes):
        self.controller_notes.extend(controller_notes)
        self.update_state_from_controller_notes()

    def add_controller_note(self, controller_note):
        self.controller_notes.append(controller_note)
        self.update_state_from_controller_notes()

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if state not in self.STATES:
            raise ValueError('"state" in ControlReport is "{}" should be one of the following : '
                             '{}'.format(str(state), ', '.join(self.STATES)))
        self._state = state

    def update_state_from_controller_notes(self):
        if len(self.controller_notes) == 0:
            self.state = self.NONE
        elif any([cn.state == ControllerNote.ERROR_RECOVERABLE for cn in self.controller_notes]):
            self.state = self.RECOVERABLE
        elif any([cn.state == ControllerNote.ERROR_UNRECOVERABLE for cn in self.controller_notes]):
            self.state = self.UNRECOVERABLE
        elif (len([cn for cn in self.controller_notes if cn.controller.can_validate]) and
                  all([cn.is_valid for cn in self.controller_notes if cn.controller.can_validate])):
            self.state = self.FINALIZED
        else:
            # If no controller says it is recoverable/unrecoverable/valid/... then we set it to unrecoverable ?
            self.state = self.UNRECOVERABLE

    @property
    def finalized(self):
        return self.state == self.FINALIZED

    @property
    def unrecoverable(self):
        return self.state == self.UNRECOVERABLE

    @property
    def actions(self):
        # TODO: should check whether actions are compatible here ... right now, two different controllers are not
        #       allowed to modify the same object ...
        actions = {}
        for cn in self.controller_notes:
            for target, action in cn.actions.items():
                if target in actions:
                    raise ValueError('Two controllers give actions to the same object!')
                actions[target] = action
        return actions

    @property
    def restart_info(self):
        if any([cn.restart == ControllerNote.RESTART_FROM_SCRATCH for cn in self.controller_notes]):
            return ControllerNote.RESTART_FROM_SCRATCH
        if any([cn.restart == ControllerNote.RESET_RESTART for cn in self.controller_notes]):
            return ControllerNote.RESET_RESTART
        return ControllerNote.SIMPLE_RESTART

    @classmethod
    def from_dict(cls, d):
        return cls(controller_notes=[ControllerNote.from_dict(cndict) for cndict in d['controller_notes']])

    def as_dict(self):
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'controller_notes': [cn.as_dict() for cn in self.controller_notes]}


#TODO: should this be MSONable ? Is that even possible with a callable object in self ?
#class Instruction(MSONable):
#class Directive(MSONable):
class Action(MSONable):

    def __init__(self, callable, **kwargs):
        self.callable = callable
        self.kwargs = kwargs

    def apply(self, object):
        if len(self.kwargs) == 0:
            self.callable(object)
        else:
            self.callable(object, **self.kwargs)

    @classmethod
    def from_dict(cls, d):
        import importlib
        mod = importlib.import_module(d['callable']['module'])
        class_ = getattr(mod, d['callable']['class'])
        callable = getattr(class_, d['callable']['func_name'])
        return cls(callable=callable)

    def as_dict(self):
        #FIXME this part is not compatible with python3 and unbound methods do not exist anymore
        return {'@class': self.__class__.__name__,
                '@module': self.__class__.__module__,
                'callable': {'module': self.callable.im_class.__module__,
                             'class': self.callable.im_class.__name__,
                             'func_name': self.callable.im_func.func_name}
                }

    @classmethod
    def from_string(cls, callable_string, **kwargs):
        #TODO: do this ?
        import importlib
        mod_name, func_name = callable_string.rsplit('.',1)
        mod = importlib.import_module(mod_name)
        callable = getattr(mod, func_name)
        cls(callable=callable, **kwargs)


from __future__ import print_function, division, unicode_literals, absolute_import

import os
import unittest
import filecmp
import tempfile
import abipy.data as abidata
import abipy.abilab as abilab

from datetime import datetime
from abipy.abio.factories import scf_input
from abiflows.core.testing import AbiflowsTest, has_mongodb, TESTDB_NAME
from abiflows.database.mongoengine.mixins import *
from mongoengine import Document
from mongoengine.errors import ValidationError
from fireworks import Workflow, Firework
from fireworks.core.rocket_launcher import rapidfire
from fireworks.user_objects.firetasks.script_task import PyTask


MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


class TestMixins(AbiflowsTest):
    """
    For all the classes, try to create an instance and call methods to fill the document.
    If mongodb is available, save the document.
    """

    @classmethod
    def setUpClass(cls):
        cls.si_structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
        cls.scf_inp = scf_input(cls.si_structure, abidata.pseudos("14si.pspnc"), ecut=2, kppa=10)

    def setUp(self):
        self.setup_mongoengine()
        self.setup_fireworks()

    @classmethod
    def tearDownClass(cls):
        cls.teardown_fireworks(module_dir=MODULE_DIR)
        cls.teardown_mongoengine()

    def test_calculation_metadata(self):
        doc_class = self.get_document_class_from_mixin(CalculationMetadataMixin)
        doc = doc_class(user="user", cluster="cluster", execution_date=datetime(2010, 1, 1, 0, 1, 1))

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_run_stats(self):
        doc_class = self.get_document_class_from_mixin(RunStatsMixin)
        doc = doc_class(core_num=10, elapsed_time=10.5, maximum_memory_used=1000, number_of_restarts=2,
                        number_of_errors=1, number_of_warnings=2)

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_spacegroup(self):
        doc_class = self.get_document_class_from_mixin(SpaceGroupMixin)
        doc = doc_class()
        doc.set_space_group_from_structure(self.si_structure)
        assert doc.number == 227

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_material(self):
        doc_class = self.get_document_class_from_mixin(MaterialMixin)
        doc = doc_class()
        doc.set_material_data_from_structure(self.si_structure)
        assert doc.nelements == 1

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_calculation_type(self):
        doc_class = self.get_document_class_from_mixin(CalculationTypeMixin)
        doc = doc_class(xc_functional="PBE", pseudo_type="nc", is_hubbard=False, pseudo_dojo_table="table")

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_ground_state_output(self):
        doc_class = self.get_document_class_from_mixin(GroundStateOutputMixin)
        doc = doc_class(final_energy=1.1, efermi=0.1, total_magnetization=-0.1, structure=self.si_structure.as_dict())

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_hubbards(self):
        doc_class = self.get_document_class_from_mixin(HubbardMixin)
        doc = doc_class(hubbards={"AA": 1.5})
        with self.assertRaises(ValidationError):
            doc.validate()

        doc = doc_class(hubbards={"Ti": 1.5})

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_date(self):
        doc_class = self.get_document_class_from_mixin(DateMixin)
        doc = doc_class(created_on=datetime(2010, 1, 1, 0, 1, 1), modified_on=datetime(2011, 1, 1, 0, 1, 1))

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1

    def test_directory(self):
        doc_class = self.get_document_class_from_mixin(DirectoryMixin)
        doc = doc_class()

        # create and run a fireworks workflow
        task = PyTask(func="time.sleep", args=[0.5])
        wf = Workflow([Firework(task, spec={'wf_task_index': "1"}, fw_id=1),
                       Firework(task, spec={'wf_task_index': "2"}, fw_id=2)])
        self.lp.add_wf(wf)
        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR)
        wf = self.lp.get_wf_by_fw_id(1)

        doc.set_dir_names_from_fws_wf(wf)
        assert len(doc.dir_names) == 2

        if has_mongodb():
            doc.save()
            query_result = doc_class.objects()
            assert len(query_result) == 1
            assert len(query_result[0].dir_names) == 2



# coding: utf-8
"""
List of mixins to provide standard interfaces with the result databases.
Most of them should still be considered as examples.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

from monty.dev import deprecated
from mongoengine import *
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.periodic_table import Element
from abiflows.core.models import AbiFileField, MSONField


class CalculationMetadataMixin(object):
    """
    Mixin providing fields for the metadata of the calculation
    """

    user = StringField(help_text="Username of the author of the calculation")
    cluster = StringField(help_text="Cluster where the calculation was performed")
    execution_date = DateTimeField(help_text="Date of the conclusion of the calculation")


class RunStatsMixin(object):
    """
    Mixin providing fields for the statistics of execution of the calculation
    """

    core_num = IntField(help_text="Number of cores used to run the calculation")
    elapsed_time = FloatField(help_text="Time in seconds required to perform the calculation")
    maximum_memory_used = FloatField(help_text="Maximum memory used during the calculation in kb")
    number_of_restarts = IntField(help_text="Number of times the calculation has been restarted")
    number_of_errors = IntField(help_text="Number of errors corrected by error handlers")
    number_of_warnings = IntField(help_text="Number of warnings in the last run")


class SpaceGroupMixin(object):
    """
    Mixin providing fields to save space group data.
    """

    crystal_system = StringField()
    hall = StringField()
    number = IntField()
    point_group = StringField()
    source = StringField()
    symbol = StringField()
    #TODO tolerances?

    @deprecated(message="fill_from_structure has been renamed set_space_group_from_structure.")
    def fill_from_structure(self, structure):
        self.set_space_group_from_structure(structure)

    def set_space_group_from_structure(self, structure):
        spga = SpacegroupAnalyzer(structure=structure)
        self.crystal_system = spga.get_crystal_system()
        self.hall = spga.get_hall()
        self.number = spga.get_space_group_number()
        self.source = "spglib"
        self.symbol = spga.get_space_group_symbol()


class SpaceGroupDocument(SpaceGroupMixin, EmbeddedDocument):
    """
    Embedded document to describe the spacegroup of a material
    """
    pass


class MaterialMixin(object):
    """
    Mixin providing the fields describing the material examined in the calculation
    """

    pretty_formula = StringField(help_text="A nice string formula where the element amounts are normalized")
    full_formula = StringField(help_text="A string of the full explicit formula for the unit cell")
    spacegroup = EmbeddedDocumentField(SpaceGroupDocument)
    unit_cell_formula = DictField(help_text="A dict of the full (unnormalized) formula of the unit cell")
    reduced_cell_formula = DictField(help_text="A dict of the normalized cell formula")
    anonymous_formula = StringField(help_text="A string of the anonymous formula for the material")
    elements = ListField(StringField(), help_text="A list of element symbols present in this material")
    nelements = IntField(help_text="Number of elements in this material's chemical system")
    nsites = IntField(help_text="Number of sites in the structure")
    chemsys = StringField(help_text="A string chemical system associated with the material")

    def set_material_data_from_structure(self, structure, space_group=True, symprec=1e-3, angle_tolerance=5):
        """
        Sets the fields of the Document using a Structure and Spglib to determine the space group properties

        Args:
            structure: A |Structure|
            space_group: if True sets the spacegroup fields using spglib_.
            symprec (float): Tolerance for symmetry finding.
            angle_tolerance (float): Angle tolerance for symmetry finding.
        """

        comp = structure.composition
        el_amt = structure.composition.get_el_amt_dict()
        self.unit_cell_formula = comp.as_dict()
        self.reduced_cell_formula = comp.to_reduced_dict
        self.elements = list(el_amt.keys())
        self.nelements = len(el_amt)
        self.pretty_formula = comp.reduced_formula
        self.anonymous_formula = comp.anonymized_formula
        self.nsites = comp.num_atoms
        self.chemsys = "-".join(sorted(el_amt.keys()))
        if space_group:
            sym = SpacegroupAnalyzer(structure, symprec=symprec, angle_tolerance=angle_tolerance)
            self.spacegroup = SpaceGroupDocument(crystal_system=sym.get_crystal_system(), hall=sym.get_hall(),
                                                 number=sym.get_space_group_number(), point_group=sym.get_point_group_symbol(),
                                                 symbol=sym.get_space_group_symbol(), source="spglib")


class CalculationTypeMixin(object):
    """
    Mixin providing fields for basic details of the calculation
    """

    xc_functional = StringField(help_text="Specific exchange-correlation functional used for the calculation")
    pseudo_type = StringField(help_text="Type of the pseudopotential", choices=["nc", "us", "paw"])
    is_hubbard = BooleanField(help_text="True if calculation was performed with hubbard correction")
    pseudo_dojo_table = StringField(help_text="String describing the version of pseudo dojo table")
    # other possibile fields: magnetization_type (unpolarized, polarized, non collinear), has_spin_orbit


class GroundStateOutputMixin(object):
    """
    Mixin providing generic fields for ground state calculation
    """

    #TODO would it be convenient to make these FloatWithUnitField?
    final_energy = FloatField(help_text="Final energy obtained computed in eV")
    efermi = FloatField(help_text="Computed Fermi energy in eV")
    total_magnetization = FloatField(help_text="Total magnetization computed") # does pymatgen have a default unit for this?

    structure = MSONField(required=True, help_text="The structure used for the calculation. If a relaxation the final structure.")


class HubbardsField(DictField):
    """
    Dictfield to validate the passed values
    """

    def validate(self, value):
        if not all(Element.is_valid_symbol(k) for k in value.keys()):
            self.error('Keys should be element symbols')
        if not all(isinstance(v, (float, int)) for v in value.values()):
            self.error('Values should be numbers')
        super(DictField, self).validate(value)


class HubbardMixin(object):
    """
    Mixin providing hubbard fields
    """
    hubbards = HubbardsField(help_text="Dict containing the values of the Hubbard U values for each element")


class DateMixin(object):
    """
    Mixin providing timestamp fields
    """

    created_on = DateTimeField()
    modified_on = DateTimeField()


class DirectoryMixin(object):
    """
    Mixin providing the field to store the directories used during the workflow
    """

    dir_names = DictField(help_text="Keys are the folder containing the various steps of the calculation, with a "
                                    "description of the step as value. Task name as value is the best option. "
                                    "None is accettable")

    def set_dir_names_from_fws_wf(self, wf):
        """
        Fills the the dir_names of the mixin using the COMPLETED launches of a firework workflow.
        Needs access to the full data of the workflow: Fireworks and Launches.
        """

        #FIXME should the imports be moved to the top? Would it be better to not rely on external objects/functions?
        from abiflows.fireworks.utils.fw_utils import get_last_completed_launch

        d = {}

        for fw in wf.fws:
            task_index = fw.spec.get('wf_task_index', None)
            if task_index is None and "unnamed" not in fw.name.lower():
                task_index = fw.name

            launch = get_last_completed_launch(fw)

            # skip the fws that have not run yet
            if launch is None:
                continue

            d[launch.launch_dir] = task_index

        self.dir_names = d


class CustomFieldMixin(object):
    """
    Mixin providing a DictField for storing user specific additional properties
    """

    custom = DictField(help_text="Dict storing additional custom properties to identify the calculation.")



"""Configuration file for pytest."""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import pytest
import abipy.abilab as abilab
import abipy.data as abidata

from pymongo import MongoClient
from abipy.abio.factories import ebands_input, ion_ioncell_relax_input, scf_for_phonons, phonons_from_gsinput
from abipy.data.benchmark_structures import simple_semiconductors, simple_metals
from abiflows.database.mongoengine.utils import DatabaseData
from fireworks import LaunchPad, FWorker
try:
    from pymatgen.ext.matproj import MPRester
except ImportError:
    from pymatgen.matproj.rest import MPRester
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


TESTDB_NAME = 'abiflows_unittest'
TESTCOLLECTION_NAME = 'test_results'
MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope="module")
def lp(request):
    lp = LaunchPad(name=TESTDB_NAME, strm_lvl='ERROR')
    lp.reset(password=None, require_password=False)

    def fin():
        lp.connection.drop_database(TESTDB_NAME)

    # request.addfinalizer(fin)
    return lp


@pytest.fixture(scope="module")
def fworker():
    return FWorker()


@pytest.fixture(scope="function")
def cleandb(request, lp):
    def fin():
        lp.reset(password=None, require_password=False)
    request.addfinalizer(fin)


@pytest.fixture(params=[False, True])
def use_autoparal(request):
    """
    This fixture allows to run some of the test both with autoparal True and False
    """
    return request.param


def get_si_structure():
    cif_file = abidata.cif_file("si.cif")
    structure = abilab.Structure.from_file(cif_file)

    return structure

def get_gan_structure():
    cif_file = abidata.cif_file("gan.cif")
    structure = abilab.Structure.from_file(cif_file)

    return structure


@pytest.fixture(scope="function")
def input_scf_si_low():
    pseudos = abidata.pseudos("14si.pspnc")
    structure = get_si_structure()

    return ebands_input(structure, pseudos, kppa=100, ecut=6, spin_mode="unpolarized",
                        accuracy="low").split_datasets()[0]

@pytest.fixture(scope="function")
def input_ebands_si_low():
    pseudos = abidata.pseudos("14si.pspnc")
    structure = get_si_structure()

    return ebands_input(structure, pseudos, kppa=100, ecut=6, ndivsm=3, spin_mode="unpolarized",
                        accuracy="low").split_datasets()


@pytest.fixture(scope="function")
def inputs_relax_si_low():
    pseudos = abidata.pseudos("14si.pspnc")
    structure = get_si_structure()

    # negative strain. This will allow to trigger dilatmx error
    structure.apply_strain(-0.005)
    structure.translate_sites(0, [0.001, -0.003, 0.005])
    structure.translate_sites(1, [0.007, 0.006, -0.005])
    return ion_ioncell_relax_input(structure, pseudos, kppa=100, ecut=4, spin_mode="unpolarized",
                                   accuracy="low", smearing=None).split_datasets()


@pytest.fixture(scope="function")
def input_scf_phonon_si_low():
    pseudos = abidata.pseudos("14si.pspnc")
    structure = get_si_structure()

    scf_in = scf_for_phonons(structure, pseudos, kppa=100, ecut=4, spin_mode="unpolarized", accuracy="low",
                             smearing=None)

    return  scf_in

@pytest.fixture(scope="function")
def input_scf_phonon_gan_low():
    pseudos = [abidata.pseudos("31ga.pspnc").pseudo_with_symbol('Ga'),
               abidata.pseudos("7n.pspnc").pseudo_with_symbol('N')]
    structure = get_gan_structure()

    scf_in = scf_for_phonons(structure, pseudos, kppa=100, ecut=4, spin_mode="unpolarized", accuracy="low",
                             smearing=None)

    return  scf_in

@pytest.fixture(scope="function")
def db_data():
    """
    Creates an instance of DatabaseData with TESTDB_NAME and TESTCOLLECTION_NAME.
    Drops the collection.
    """
    db_data = DatabaseData(TESTDB_NAME, collection=TESTCOLLECTION_NAME)
    connection = db_data.connect_mongoengine()
    connection[TESTDB_NAME].drop_collection(TESTCOLLECTION_NAME)

    return db_data


@pytest.fixture(scope="function", params=simple_semiconductors)
def benchmark_input_scf(request):
    pseudos = abidata.pseudos("14si.pspnc", "6c.pspnc", "3li.pspnc", "9f.pspnc",
                              "12mg.pspnc", "8o.pspnc", "31ga.pspnc", "7n.pspnc")
    rest = MPRester()
    structure = rest.get_structure_by_material_id(request.param)
    try:
        return ebands_input(structure, pseudos, kppa=100, ecut=6).split_datasets()[0]
    except:
        #to deal with missing pseudos
        pytest.skip('Cannot create input for material {}.'.format(request.param))


@pytest.fixture()
def fwp(tmpdir):
    """
    Parameters used to initialize Flows.
    """
    # Temporary working directory
    fwp.workdir = str(tmpdir)

    # Create the TaskManager.
    fwp.manager = abilab.TaskManager.from_file(os.path.join(os.path.dirname(__file__), "manager.yml"))

    fwp.scheduler = abilab.PyFlowScheduler.from_file(os.path.join(os.path.dirname(__file__), "scheduler.yml"))

    return fwp






# coding: utf-8
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import shutil
import unittest
import yaml

from monty.tempfile import ScratchDir
from fireworks import Firework, LaunchPad
from fireworks.core.rocket_launcher import rapidfire
from fireworks.user_objects.firetasks.script_task import PyTask
from abiflows.core.testing import AbiflowsTest, has_mongodb, TESTDB_NAME
from abiflows.fireworks.utils.fw_utils import *
from abiflows.fireworks.utils.tests.tasks import LpTask


MODULE_DIR = os.path.dirname(os.path.abspath(__file__))

test_dir = os.path.join(os.path.dirname(__file__), "..", "..", "..", "..",
                        "test_files", "fw_task_managers")



class TestFWTaskManager(AbiflowsTest):

    def test_abipy_manager_from_file(self):
        with open(os.path.join(test_dir, "fw_manager_ok.yaml"), "rt") as fh:
            conf = yaml.load(fh)
        conf['fw_policy']['abipy_manager'] = os.path.join(test_dir, "manager_ok.yml")
        ftm = FWTaskManager(**conf)

        self.assertEqual(ftm.fw_policy.max_restarts, 20)

        ftm.update_fw_policy({'max_restarts': 30})

        self.assertTrue(ftm.has_task_manager())
        self.assertTrue(ftm.fw_policy.rerun_same_dir)
        self.assertEqual(ftm.fw_policy.max_restarts, 30)
        self.assertTrue(ftm.fw_policy.autoparal)

    def test_ok(self):

        ftm = FWTaskManager.from_file(os.path.join(test_dir, "fw_manager_ok.yaml"))
        ftm.update_fw_policy({'max_restarts': 30})

        self.assertTrue(ftm.fw_policy.rerun_same_dir)
        self.assertEqual(ftm.fw_policy.max_restarts, 30)
        self.assertTrue(ftm.fw_policy.autoparal)

    def test_local_qadapter(self):

        ftm = FWTaskManager.from_file(os.path.join(test_dir, "fw_manager_local_qadapter.yaml"))
        ftm.update_fw_policy({'max_restarts': 30})

        self.assertTrue(ftm.fw_policy.rerun_same_dir)
        self.assertEqual(ftm.fw_policy.max_restarts, 30)
        self.assertTrue(ftm.fw_policy.autoparal)

    def test_unknown_keys(self):

        with self.assertRaises(RuntimeError):
            ftm = FWTaskManager.from_file(os.path.join(test_dir, "fw_manager_unknown_keys.yaml"))

    def test_from_user_config(self):

        # create also using the from_user_config classmethod. Copy the file in the current folder
        with ScratchDir("."):
            shutil.copy2(os.path.join(test_dir, "fw_manager_ok.yaml"),
                         os.path.join(os.getcwd(), FWTaskManager.YAML_FILE))
            ftm = FWTaskManager.from_user_config()
            ftm.update_fw_policy({'max_restarts': 30})

            self.assertTrue(ftm.fw_policy.rerun_same_dir)
            self.assertEqual(ftm.fw_policy.max_restarts, 30)
            self.assertTrue(ftm.fw_policy.autoparal)


class TestFunctions(AbiflowsTest):

    @classmethod
    def setUpClass(cls):
        cls.setup_fireworks()
        ftm_path = os.path.join(test_dir, "fw_manager_ok.yaml")
        with open(ftm_path, "rt") as fh:
            conf = yaml.load(fh)
        conf['fw_policy']['abipy_manager'] = os.path.join(test_dir, "manager_ok.yml")
        cls.ftm = FWTaskManager(**conf)

    @classmethod
    def tearDownClass(cls):
        cls.teardown_fireworks(module_dir=MODULE_DIR)

    def tearDown(self):
        if self.lp:
            self.lp.reset(password=None,require_password=False)

    def test_get_short_single_core_spec(self):
        spec = get_short_single_core_spec(self.ftm, timelimit=610)

        assert spec['ntasks'] == 1
        assert spec['time'] == '0-0:10:10'

    def test_set_short_single_core_to_spec(self):
        spec = {}
        spec = set_short_single_core_to_spec(spec, fw_manager=self.ftm)

        assert spec['_queueadapter']['ntasks'] == 1
        assert spec['mpi_ncpus'] == 1

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_get_time_report_for_wf(self):
        task = PyTask(func="time.sleep", args=[0.5])
        fw1 = Firework([task], spec={'wf_task_index': "test1_1", "nproc": 16}, fw_id=1)
        fw2 = Firework([task], spec={'wf_task_index': "test2_1", "nproc": 16}, fw_id=2)
        wf = Workflow([fw1,fw2])
        self.lp.add_wf(wf)

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR)

        wf = self.lp.get_wf_by_fw_id(1)

        assert wf.state == "COMPLETED"

        tr = get_time_report_for_wf(wf)

        assert tr.n_fws == 2
        assert tr.total_run_time > 1

    @unittest.skipUnless(has_mongodb(), "A local mongodb is required.")
    def test_get_lp_and_fw_id_from_task(self):
        """
        Tests the get_lp_and_fw_id_from_task. This test relies on the fact that the LaunchPad loaded from auto_load
        will be different from what is defined in TESTDB_NAME. If this is not the case the test will be skipped.
        """
        lp = LaunchPad.auto_load()

        if not lp or lp.db.name == TESTDB_NAME:
            raise unittest.SkipTest("LaunchPad lp {} is not suitable for this test. Should be available and different"
                                    "from {}".format(lp, TESTDB_NAME))

        task = LpTask()
        # this will pass the lp
        fw1 = Firework([task], spec={'_add_launchpad_and_fw_id': True}, fw_id=1)
        # this will not have the lp and should fail
        fw2 = Firework([task], spec={}, fw_id=2, parents=[fw1])
        wf = Workflow([fw1, fw2])
        self.lp.add_wf(wf)

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR, nlaunches=1)

        fw = self.lp.get_fw_by_id(1)

        assert fw.state == "COMPLETED"

        rapidfire(self.lp, self.fworker, m_dir=MODULE_DIR, nlaunches=1)

        fw = self.lp.get_fw_by_id(2)

        assert fw.state == "FIZZLED"


from __future__ import print_function, division, unicode_literals, absolute_import

from abiflows.database.mongoengine.mixins import MaterialMixin, DateMixin, DirectoryMixin, CustomFieldMixin
from abiflows.database.mongoengine.abinit_mixins import AbinitPhononOutputMixin, AbinitBasicInputMixin, \
    AbinitGSOutputMixin, AbinitPhononOutputMixin, AbinitDftpOutputMixin
from mongoengine import *
from abiflows.core.models import AbiFileField, MSONField, AbiGzipFileField

class RelaxAbinitInput(AbinitBasicInputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical inputs for a relaxation workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RelaxAbinitInput
    """

    last_input = MSONField(help_text="The last input used for the calculation.")
    kppa = IntField()


class RelaxAbinitOutput(AbinitGSOutputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical outputs for a relaxation workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RelaxAbinitOutput
    """

    hist_files =  DictField(field=AbiFileField(abiext="HIST.nc", abiform="b", collection_name='relax_hist_fs'),
                            help_text="Series of HIST files produced during the relaxation. Keys should provide the type"
                                      " of relaxation (ion, ioncell) and the ordering", db_field='relax_hist_files_ids')
    outfile_ioncell = AbiGzipFileField(abiext="abo", abiform="t", db_field='outfile_ioncell_id',
                                       collection_name='relax_outfile_fs')


class RelaxResult(MaterialMixin, DateMixin, DirectoryMixin, CustomFieldMixin, Document):
    """
    Document containing the results for a relaxation workflow consisting of an ion followed by ioncell relaxations.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: RelaxResult
    """

    history = DictField()
    mp_id = StringField()
    abinit_input = EmbeddedDocumentField(RelaxAbinitInput, default=RelaxAbinitInput)
    abinit_output = EmbeddedDocumentField(RelaxAbinitOutput, default=RelaxAbinitOutput)
    time_report = MSONField()
    fw_id = IntField()


class PhononAbinitInput(AbinitBasicInputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical inputs for a phonon workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PhononAbinitInput
    """

    gs_input = MSONField(help_text="The last input used to calculate the wafunctions.")
    ddk_input = MSONField(help_text="The last input used to calculate one of the ddk.")
    dde_input = MSONField(help_text="The last input used to calculate one of the dde.")
    wfq_input = MSONField(help_text="The last input used to calculate one of the wfq.")
    phonon_input = MSONField(help_text="The last input used to calculate one of the phonons.")
    kppa = IntField()
    ngqpt = ListField(IntField())
    qpoints = ListField()
    qppa = IntField()


class PhononAbinitOutput(AbinitPhononOutputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical outputs for a phonon workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PhononAbinitOutput
    """

    gs_gsr = AbiFileField(abiext="GSR.nc", abiform="b", help_text="Gsr file produced by the Ground state calculation",
                          db_field='gs_gsr_id', collection_name='phonon_gs_gsr_fs')
    gs_outfile= AbiGzipFileField(abiext="abo", abiform="t", db_field='gs_outfile_id',
                                 collection_name='phonon_gs_outfile_fs')


class PhononResult(MaterialMixin, DateMixin, DirectoryMixin, CustomFieldMixin, Document):
    """
    Document containing the results for a phonon workflow.
    Includes information from the various steps of the workflow (scf, nscf, ddk, dde, ph, anaddb)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PhononResult
    """

    mp_id = StringField()
    # relax_result = ReferenceField(RelaxationResult)
    relax_db = MSONField()
    relax_id = StringField()
    time_report = MSONField()
    fw_id = IntField()
    abinit_input = EmbeddedDocumentField(PhononAbinitInput, default=PhononAbinitInput)
    abinit_output = EmbeddedDocumentField(PhononAbinitOutput, default=PhononAbinitOutput)


class DteAbinitInput(AbinitBasicInputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical inputs for a DTE workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DteAbinitInput
    """

    gs_input = MSONField(help_text="The last input used to calculate the wafunctions.")
    ddk_input = MSONField(help_text="The last input used to calculate one of the ddk.")
    dde_input = MSONField(help_text="The last input used to calculate one of the dde.")
    dte_input = MSONField(help_text="The last input used to calculate one of the dte.")
    phonon_input = MSONField(help_text="The last input used to calculate one of the phonons.")
    kppa = IntField()
    with_phonons = BooleanField(help_text="Whether the phonon perturbations have been included or not")


class DteAbinitOutput(AbinitDftpOutputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical outputs for a DTE workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DteAbinitOutput
    """

    gs_gsr = AbiFileField(abiext="GSR.nc", abiform="b", help_text="Gsr file produced by the Ground state calculation",
                          db_field='gs_gsr_id', collection_name='phonon_gs_gsr_fs')
    gs_outfile= AbiGzipFileField(abiext="abo", abiform="t", db_field='gs_outfile_id',
                                 collection_name='phonon_gs_outfile_fs')
    anaddb_nc = AbiFileField(abiext="anaddb.nc", abiform="b", db_field='anaddb_nc_id', collection_name='anaddb_nc_fs')

    epsinf = ListField(ListField(FloatField()), help_text="Macroscopic dielectric tensor")
    eps0 = ListField(ListField(FloatField()), help_text="Relaxed ion Macroscopic dielectric tensor")
    dchide= ListField(ListField(ListField(FloatField())), help_text="Non-linear optical susceptibilities tensor")
    dchidt= ListField(ListField(ListField(ListField(FloatField()))),
                      help_text="First-order change in the linear dielectric susceptibility")


class DteResult(MaterialMixin, DateMixin, DirectoryMixin, CustomFieldMixin, Document):
    """
    Document containing the results for a dte workflow.
    Includes information from the various steps of the workflow (scf, ddk, dde, ph, dte, anaddb)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DteResult
    """

    mp_id = StringField()
    relax_db = MSONField()
    relax_id = StringField()
    time_report = MSONField()
    fw_id = IntField()
    abinit_input = EmbeddedDocumentField(DteAbinitInput, default=DteAbinitInput)
    abinit_output = EmbeddedDocumentField(DteAbinitOutput, default=DteAbinitOutput)


class DfptAbinitInput(AbinitBasicInputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical inputs for a dfpt workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DfptAbinitInput
    """

    gs_input = MSONField(help_text="The last input used to calculate the wafunctions.")
    ddk_input = MSONField(help_text="The last input used to calculate one of the ddk.")
    dde_input = MSONField(help_text="The last input used to calculate one of the dde.")
    wfq_input = MSONField(help_text="The last input used to calculate one of the wfq.")
    phonon_input = MSONField(help_text="The last input used to calculate one of the phonons.")
    strain_input = MSONField(help_text="The last input used to calculate one of the strain.")
    dte_input = MSONField(help_text="The last input used to calculate one of the dte.")
    kppa = IntField()
    ngqpt = ListField(IntField())
    qpoints = ListField()
    qppa = IntField()


class DfptAbinitOutput(AbinitPhononOutputMixin, EmbeddedDocument):
    """
    EmbeddedDocument containing the typical outputs for a dfpt workflow.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DfptAbinitOutput
    """

    gs_gsr = AbiFileField(abiext="GSR.nc", abiform="b", help_text="Gsr file produced by the Ground state calculation",
                          db_field='gs_gsr_id', collection_name='phonon_gs_gsr_fs')
    gs_outfile= AbiGzipFileField(abiext="abo", abiform="t", db_field='gs_outfile_id',
                                 collection_name='phonon_gs_outfile_fs')


class DfptResult(MaterialMixin, DateMixin, DirectoryMixin, CustomFieldMixin, Document):
    """
    Document containing the results for a dfpt workflow.
    Includes information from the various steps of the workflow (scf, nscf, ddk, dde, ph, strain, dte, anaddb)

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: DfptResult
    """

    mp_id = StringField()
    relax_db = MSONField()
    relax_id = StringField()
    time_report = MSONField()
    fw_id = IntField()
    abinit_input = EmbeddedDocumentField(DfptAbinitInput, default=DfptAbinitInput)
    abinit_output = EmbeddedDocumentField(DfptAbinitOutput, default=DfptAbinitOutput)

    has_phonons = BooleanField(help_text="Whether the phonon perturbations have been included or not", default=False)
    has_ddk = BooleanField(help_text="Whether the ddk perturbations have been included or not", default=False)
    has_dde = BooleanField(help_text="Whether the dde perturbations have been included or not", default=False)
    has_strain = BooleanField(help_text="Whether the strain perturbations have been included or not", default=False)
    has_dte = BooleanField(help_text="Whether the dte perturbations have been included or not", default=False)



from __future__ import print_function, division, unicode_literals, absolute_import

from math import sqrt
from six.moves import reduce

##############################################################
### cartesian product of lists ##################################
##############################################################

def _append_es2sequences(sequences, es):
    result = []
    if not sequences:
        for e in es:
            result.append([e])
    else:
        for e in es:
            result += [seq+[e] for seq in sequences]
    return result


def _cartesian_product(lists):
    """
    given a list of lists,
    returns all the possible combinations taking one element from each list
    The list does not have to be of equal length
    """
    return reduce(_append_es2sequences, lists, [])


def prime_factors(n):
    """Lists prime factors of a given natural integer, from greatest to smallest
    :param n: Natural integer
    :rtype : list of all prime factors of the given natural n
    """
    i = 2
    while i <= sqrt(n):
        if n % i == 0:
            l = prime_factors(n/i)
            l.append(i)
            return l
        i += 1
    return [n]      # n is prime


def _factor_generator(n):
    """
    From a given natural integer, returns the prime factors and their multiplicity
    :param n: Natural integer
    :return:
    """
    p = prime_factors(n)
    factors = {}
    for p1 in p:
        try:
            factors[p1] += 1
        except KeyError:
            factors[p1] = 1
    return factors


def divisors(n):
    """
    From a given natural integer, returns the list of divisors in ascending order
    :param n: Natural integer
    :return: List of divisors of n in ascending order
    """
    factors = _factor_generator(n)
    _divisors = []
    listexponents = [[k**x for x in range(0, factors[k]+1)] for k in list(factors.keys())]
    listfactors = _cartesian_product(listexponents)
    for f in listfactors:
        _divisors.append(reduce(lambda x, y: x*y, f, 1))
    _divisors.sort()
    return _divisors